#SkIndigo Ruby code is copyright Dale Martens 2009, all rights reserved.

######################
module SkIndigo

	#method to collect all indigo instances from a collection of entities
	def SkIndigo.get_instances(ents)

		instances=ents.find_all {|e|
			((e.class==Sketchup::ComponentInstance) or (e.class==Sketchup::Group))
			#if e.class==Sketchup::ComponentInstance
			#	e.definition.indigo_instance?
			#elsif e.class==Sketchup::Group
			#	e.entities.parent.indigo_instance?
			#end
		}
		return instances
	end

	#interface for settings motion blur keyframes
	def SkIndigo.set_keyframe(ents)

		return if ents.length==0
		prompts=["Time (0.0-1.0)"]
		values=[0.0]
		results=UI.inputbox(prompts,values,"Set Keyframes")
		return if not results
		time=results[0].to_f
		if time>1.0 or time<0.0
			UI.messagebox("Time must be between 0.0 and 1.0")
			return
		end
		instances=self.get_instances(ents)
		if instances.length==0
			UI.messagebox("No instances selected.  No keyframes added.")
		else
			instances.each {|inst| inst.add_indigo_keyframe(time)}
		end
		UI.messagebox("Added keyframe for #{instances.length} instances.")
	end

	######clear all keyfreames
	def SkIndigo.clear_keyframes(ents)

		instances=self.get_instances(ents)
		if instances.length==0
			UI.messagebox("No instances selected.")
		else
			instances.each {|inst| inst.clear_indigo_keyframes}
		end
		UI.messagebox("Cleared keyframes for #{instances.length} instances.")
	end
		

	def smoothstep(min,max,input) 
		r=(input.to_f-min.to_f)/(max.to_f-min.to_f)
		return r*r*(3.0-2.0*r)
	end

	#Check for saved sketchy physics animation in the file
	def SkIndigo.saved_sketchy_replay?
		sr=SketchyReplay::SketchyReplay.new()
		if sr.lastFrame==0
			return false
		else
			return true
		end
	end
	
	
	####################################################
	def SkIndigo.export_animation(sketchyphysics)  # sketchyphysics is a flag for whether to export the sketchyphysics animation
		begin
			if IndigoRenderSettings.new.halt.to_i == -1
				UI.messagebox("Warning: Halt time is set to -1. This means that each frame will render forever unless the halt time is set to a value such as 10 (s).  Halt time can be set in the SkIndigo render settings dialog.")
			end

			if not sketchyphysics
				prompts = ["Frame Rate", "Smooth Transitions", "Animate Objects"]
				values = ["25.0", "Off", "Off"]
				enums = [nil, "On|Off", "On|Off"]
				results = UI.inputbox(prompts, values, enums, "Animation Options")
				return if not results
				smooth = true if results[1] == "On"
				smooth = false if results[1] == "Off"
				animate_objects = true if results[2] == "On"
				animate_objects = false if results[2] == "Off"
			else
				if not saved_sketchy_replay?
					UI.messagebox("SketchyReplay animation not found.  You must first save an animation using SketchyPhysics before using this feature.")
					return
				end
				# Set objects pos and camera for first frame.
				sr = SketchyReplay::SketchyReplay.new()
				sr.start()
			end
			
			print_timings = !SkIndigo.run_tests?

			model = Sketchup.active_model
			model_filename = File.basename(model.path, ".skp")
			if model_filename != ""
				queue_filename = model_filename + ".igq"
			else
				queue_filename = "Untitled.igq" 	
			end
			batch_file_path = UI.savepanel("Export Animation", "", queue_filename) # path for the IGQ file
			return if batch_file_path == nil

			export_dir = File.join(File.dirname(batch_file_path), File.basename(batch_file_path, ".*"))
			Dir.mkdir(export_dir) unless FileTest.exist?(export_dir)
			export_path = File.join(export_dir, File.basename(batch_file_path, ".*")) + ".igs" 


			ie = IndigoExporter.new(export_path)
			ie.physics = true if sketchyphysics  #set the sketchyphysics flag as this modifies mesh exporting by setting the mesh axis to the center of the bounding box
			ie.settings.save_to_model()  #get the settings from the model, not the current scene

			model = ie.model
			model_name = File.basename(ie.path, ".igs")
			return if model_name == ""
			tex_path = File.join(File.dirname(export_path), "TX_" + model_name)

			ie.tex_path = "TX_" + model_name
			
			# Open the shared IGS file to export to.
			shared_igs_path = File.join(File.dirname(ie.path), model_name + "-shared.igs")
			shared_igs_file = SkIndigo.new_writable_utf8_file(shared_igs_path) 
			shared_igs_file.puts "<?xml version=\"1.0\" encoding=\"utf-8\"?>"
			shared_igs_file.puts "<scenedata>"
			
			ie.export_default_mat(shared_igs_file)

			mesh_builder = IndigoMeshBuilder.new
			IndigoMeshBuilder.reset  # Erases the stored meshes, instances, and spheres from last export
			
			# Process entities (iterate over scene, get list of components/groups etc.. to export)
			t = Time.now
			Sketchup.set_status_text("Collecting Instances...")
			entities_to_export = Sketchup.active_model.entities
			mesh_builder.process_entities(entities_to_export)
			puts "Traversed model and got instances in #{Time.now-t} seconds."
			
			# Build required (visible, referenced) meshes
			t = Time.now
			Sketchup.set_status_text("Building Meshes...")
			mesh_builder.build_meshes()
			puts "Built meshes in #{Time.now-t} seconds."	
			
			# Write meshes
			t = Time.now
			Sketchup.set_status_text("Exporting Meshes...")
			ie.export_meshes(mesh_builder, shared_igs_file)
			puts "Exported meshes in #{Time.now-t} seconds."

			
			ie.get_used_layers()
			
			
			# Build a set of the materials that were actually referenced by a model, so that we can just export those materials, and not all materials in the scene.
			referenced_materials = ie.get_referenced_materials(mesh_builder)
			
			#----------- Export all the materials -------------
			t = Time.now
			Sketchup.set_status_text("Exporting Materials...")
			referenced_materials.each {|mat, generate_uvs|
				ie.export_material(IndigoMaterial.new(mat), shared_igs_file, generate_uvs)
			}
			puts "Exported materials in #{Time.now-t} seconds."	if print_timings
			#------------------------------------------------

			# Write instances
			if !animate_objects
				t = Time.now
				Sketchup.set_status_text("Exporting Instances...")
				
				ie.export_instances(mesh_builder, shared_igs_file)
				
				puts "Exported instances in #{Time.now-t} seconds." if print_timings
			end

			# Write scatters
			ie.export_scatters(mesh_builder, shared_igs_file)
			
			
			
			shared_igs_file.puts "</scenedata>"
			shared_igs_file.close

			#------------------Export textures---------------
			Sketchup.set_status_text("Exporting textures")		
			Dir.mkdir(tex_path) if !FileTest.exist?(tex_path)

			# Export all textures collected during the material export process
			ie.export_textures(ie.textures)
					
			#Export all collected nkata
			#for path in ie.nkdata
			#	nk=NKdata.new(path)
			#	nk.export(tex_path)  #exports the nkdata to the textures directory
			#end
			#------------------End export textures---------------

			#Export all collected ies profiles
			#p "ies profiles #{ie.ies}"
			for path in ie.ies
				profile=IESProfile.new(path)
				profile.export(tex_path) if profile.valid?  #export the IES profiles to the textures directory
			end
			
			if print_timings
				puts "------SkIndigo Export Run------"	
				puts "Instances: #{mesh_builder.instances.size}"
				puts "Meshes: #{mesh_builder.num_meshes()}"
				puts "Meshes exported: #{mesh_builder.num_new_meshes_exported}"
				puts "Spheres: #{mesh_builder.spheres.length}"
			end

			if !sketchyphysics
				ie.export_scene_tab_animation(File.dirname(export_path), results[0].to_f, smooth, animate_objects, mesh_builder, shared_igs_path)
			else #we are exporting a sketchyphysics animation
				ie.export_sketchyphysics_frame(File.dirname(export_path), mesh_builder, shared_igs_path)  #export the first frame
				 0.upto(sr.lastFrame){
					#advance object and camera positions
					sr.nextFrame()
					ie.export_sketchyphysics_frame(File.dirname(export_path), mesh_builder, shared_igs_path) 
				}
				sr.rewind()
			end
			
			ie.create_animation_igq_file(batch_file_path)
			
			#--------------Open Indigo for rendering
			is = IndigoRenderSettings.new
			
			indigo_path = SkIndigo.get_indigo_path
			if indigo_path == nil
				UI.messagebox "Indigo application not found.  Export aborted."
				return
			end
			
			auto_render = true
			if auto_render && FileTest.exist?(indigo_path)

				output_param = ""
				case is.network_mode
				when 1 # master
					output_param += "-n m " unless SkIndigo.on_mac?
				when 2 # working master
					output_param += "-n wm " unless SkIndigo.on_mac?
				end
				
				SkIndigo.launch_indigo(indigo_path, batch_file_path, output_param, is.low_priority?, "normal")		
			end
			
		rescue => exception
			UI.messagebox(
				"An error occurred while SkIndigo was exporting the animation.\n" + 
				"Please copy and paste the following details into an email to support@indigorenderer.com:\n" + 
				"\n" + 
				"SkIndigo version: " + VERSION + "\n" + 
				"SketchUp version: " + Sketchup.version.to_s + "\n" + 
				"Ruby platform: " + Object::RUBY_PLATFORM + "\n" + 
				"Error during processing: #{$!} \n" + 
				exception.backtrace.join("\n\t") + "\n",
				MB_MULTILINE, 
				"Error occurred in SkIndigo"
			)
		end # End exception rescue
	end #export_animation

end #module SkIndigo









###################
class IndigoExporter  

	include SkIndigo

	def export_scene_tab_animation(igs_path, frame_rate, smooth, animate_objects, mesh_builder, shared_igs_path)

		Dir.mkdir(igs_path) unless FileTest.exist?(igs_path) # igs_path is a directory name that stores the IGS files for each frame
		return 0 unless FileTest.exist?(igs_path)    # Failed to create path
		pages = Sketchup.active_model.pages
		#puts "pages: #{pages.inspect}"
		if pages.count == 0
			puts "pages.count == 0!"
			return 0
		end

		# Set all page delay times to 0.0 (restore after animation)
		delay_times = pages.collect {|page| page.delay_time}
		pages.each {|page| page.delay_time = 0.0}

		begin
			def_trans_time = Sketchup.active_model.options["PageOptions"]["TransitionTime"]
			total_time = pages.slideshow_time

			time_step = 1.0 / frame_rate.to_f
			num_frames = (total_time * frame_rate).to_i
			igs_path = File.basename(self.path)       
			model_name = File.basename(igs_path, ".*") # minus the file extension
			#shared_igs_path = File.join(File.dirname(self.path), model_name + "-shared.igs")
			time = 0.0

			page_start_time = time
			page_trans = pages[0].transition_time
			page_trans = def_trans_time if page_trans == -1.0
			page_end_time = page_start_time + page_trans
			@frame = num_frames
			mb = mesh_builder

			num_frames.times {|i|
			
				Sketchup.set_status_text("Exporting Animation Frame #{i}")
						
				#pages.show_frame_at(time+0.01)
				get_page = pages.show_frame_at(time)
				current_page = get_page[0]
				progress = get_page[1] # The fraction that we are into the transition
				
				
				if time >= page_end_time
					page_start_time = page_end_time
					page_trans = current_page.transition_time
					page_trans = def_trans_time if page_trans == -1.0
					page_end_time = page_start_time + page_trans
				end
				frame_num = "%04d" % i
				frame_path = File.join(File.dirname(self.path), model_name + "-#{frame_num}.igs")
				
				smoothed_time = smoothstep(0.0, 1.0, progress.to_f) * (page_end_time - page_start_time) + page_start_time
				pages.show_frame_at(smoothed_time) if smooth
				
				get_used_layers()
				
				# Open up a file for this frame's IGS
				out = SkIndigo.new_writable_utf8_file(frame_path)
				out.puts "<?xml version=\"1.0\" encoding=\"utf-8\"?>"
				out.puts "<scene>"
				
				# Delete the background material from @material_uids, so it will be re-exported.
				bg_mat = Sketchup.active_model.materials.find {|mat| mat.name == settings.bg_material}
				if bg_mat
					@material_uids.delete([bg_mat, false]) # Delete with all possibilities for generate_uvs.
					@material_uids.delete([bg_mat, true])
					@material_uids.delete([bg_mat, "Spherical"])
				end
				
				export_metadata(out)
				export_render_settings(out)
				export_tonemapping(out)
				export_environment(out)
				export_camera(model.active_view, out)
				
				mb.spheres.each {|sphere| export_sphere(sphere,out)}
				
				export_fog(out) if self.model.rendering_options["DisplayFog"]

				# Include the shared IGS file.
				out.puts "	<include>"
				out.puts "		<offset_uids>false</offset_uids>"
				out.puts "		<pathname>#{File.basename(shared_igs_path)}</pathname>" # TODO: CDATA
				out.puts "	</include>"
				out.puts ""
				
				export_instances(mb, out) if animate_objects
				
				export_light_layer_names(out)
				
				out.puts "</scene>"
				out.close
				
				time = time + time_step
			}

		ensure
			#now restore the scene delay times
			pages.count.times {|i| pages[i].delay_time = delay_times[i] }
		end
	end #export animation

	
	def export_sketchyphysics_frame(igs_path, mb, shared_igs_path)

		Dir.mkdir(igs_path) unless FileTest.exist?(igs_path)    #igs_path is a directory name that stores the IGS files for each frame - move function to initialization method???
		return 0 unless FileTest.exist?(igs_path)    #failed to create path

		@next_uid += 100

		# put this in initialization method?
		igs_path = File.basename(self.path)       
		model_name = File.basename(igs_path,".*")  #minus the file extension

		# Begin exporting the current frame
		ents = Sketchup.active_model.entities
		#mb=IndigoMeshBuilder.new
		mb.reset_instances
		#mb.build_active_mesh(ents)
		
		#exported_entity_set = Set.new() # Used to prevent infinite loops on recursive components.
		#mb.get_instances(ents,Geom::Transformation.new, exported_entity_set)
		entities_to_export = Sketchup.active_model.entities
		mb.process_entities(entities_to_export)
		#mb.build_meshes2()
		model = self.model
		
		Sketchup.set_status_text("Exporting Animation SketchyPhysics Frame #{@frame}")
		frame_num = "%04d" % @frame
		frame_path = File.join(File.dirname(self.path), model_name + "-#{frame_num}.igs")
		get_used_layers()
		out = SkIndigo.new_writable_utf8_file(frame_path)
		out.puts "<?xml version=\"1.0\" encoding=\"utf-8\"?>"
		out.puts "<scene>"
		export_metadata(out)
		export_render_settings(out)
		export_tonemapping(out)
		export_camera(model.active_view,out)
		export_environment(out)
		export_default_mat(out)
		mb.spheres.each {|sphere| export_sphere(sphere,out)}
		
		export_fog(out) if model.rendering_options["DisplayFog"]

		out.puts "	<include>"
		out.puts "		<offset_uids>false</offset_uids>"
		out.puts "		<pathname>#{File.basename(shared_igs_path)}</pathname>"
		out.puts "	</include>"
		out.puts ""
		
		export_instances(mb, out)
		export_light_layer_names(out)
		out.puts "</scene>"
		out.close
		@frame += 1
	end  #sketchyphysics frame


	def create_animation_igq_file(igq_path)

		model = self.model
		model_name = File.basename(self.path, ".igs")

		irs = IndigoRenderSettings.new

		image_name = irs.image_name
		halt = irs.halt

		indigo_path = SkIndigo.get_indigo_path
		return 0 if not indigo_path

		batch = SkIndigo.new_writable_utf8_file(igq_path)
		batch.puts "<?xml version=\"1.0\" encoding=\"utf-8\"?>"
		batch.puts "<render_queue>"

		render_output_dir = File.join(File.dirname(self.path), "renders")
		Dir.mkdir(render_output_dir) unless FileTest.exist?(render_output_dir)
		
		@frame.times {|i|
			frame_num = "%04d" % i
			export_path = File.join(File.dirname(self.path), model_name + "-#{frame_num}.igs")
			image_name  = File.join(render_output_dir, "frame-#{frame_num}")
			
			batch.puts "	<item>"
			batch.puts "		<scene_path>#{export_path}</scene_path>"
			batch.puts "		<halt_time>#{halt}</halt_time>" if irs.halt_unit == "sec"
			batch.puts "		<halt_spp>#{halt}</halt_spp>" if irs.halt_unit == "spp"
			batch.puts "		<output_path>#{image_name}</output_path>"
			batch.puts "	</item>"
		}
		batch.puts "</render_queue>"

		batch.close
	end

end  #end class indigo exporter


	
	
##############
class IndigoKeyframe
	def initialize
	end
end

	
	
##########methods for motion blur implementation
module IndigoKeyframes

	def add_indigo_keyframe(time)
		time=time.to_f
		trans=self.transformation.to_a
		keyframes=self.get_indigo_keyframes
		keyframes.push([time,trans])
		new_keyframes=self.set_attribute('indigo07_settings','keyframes',keyframes)	
	end

	def get_indigo_keyframes
		self.get_attribute('indigo07_settings','keyframes',[])
	end

	def clear_indigo_keyframes
		dict=self.attribute_dictionaries["indigo07_settings"]
		if dict
			dict.delete_key("keyframes")
		end
	end

	def has_indigo_keyframes?
		return self.get_indigo_keyframes.length>0
	end

end #module

class Sketchup::ComponentInstance
	include IndigoKeyframes
end

class Sketchup::Group
	include IndigoKeyframes
end
