#SkIndigo Ruby code is copyright Glare Technologies Limited, 2014, all rights reserved
#-----------------------------------------------------------------------------
#

class ProtoObjectGroup
	attr_accessor :prototype_def_name, :weight

	def initialize()
		@prototype_def_name = nil
		@weight = 1.0
	end
end

class IndigoScatter

	attr_accessor :target_def_name
	attr_accessor :density
	attr_accessor :seed
	attr_accessor :view_culling
	attr_accessor :orient_normal
	attr_accessor :scale_std_dev
	attr_accessor :height_std_dev
	attr_accessor :evenness
	attr_accessor :groups
	
	def initialize()
		@seed = 1
		@density = 1.0
		@view_culling = false
		@orient_normal = false
		@scale_std_dev = 0.0
		@height_std_dev = 0.0
		@evenness = 0.5
		@groups = [] # Array of ProtoObjectGroup's
	end
	
	def name
		if !@groups.empty? && @groups[0].prototype_def_name
			return @groups[0].prototype_def_name + " on " + (@target_def_name ? @target_def_name : "_______")
		else
			return "New scatter"
		end
	end
	
end



class IndigoScatterSettings

	attr_accessor :scatters

	def initialize()
		@model = Sketchup.active_model
		@dict = "indigo07_scatter_settings"
		
		@scatters = []
		
		load_scatter_settings()
	end
	
	
	def valuesChanged()
		#puts "IndigoScatterSettings: valuesChanged()"
		save_scatter_settings()
	end


	def load_scatter_settings
		#puts "load_scatter_settings()"
		
		default = []
		
		scatters_data = @model.get_attribute(@dict, "scatters", default)
		
		#puts "scatters_data: #{scatters_data.inspect}"
		
		# Convert to IndigoScatter objects
		
		@scatters = scatters_data.map do |scatter_data| 
			
			scatter = IndigoScatter.new
			
			scatter_data.each do |f|
				value = f[1]
				case f[0]
				when "prototype_def_name"
					#scatter.prototype_def_name = value
					# Backwards-compat parse
					group = ProtoObjectGroup.new
					group.prototype_def_name = value
					group.weight = 1.0
					scatter.groups = [group]
				when "target_def_name"
					scatter.target_def_name = value
				when "density"
					scatter.density = value.to_f
				when "seed"
					scatter.seed = value.to_i
				when "view_culling"
					scatter.view_culling = value
				when "orient_normal"
					scatter.orient_normal = value
				when "scale_std_dev"
					scatter.scale_std_dev = value.to_f
				when "height_std_dev"
					scatter.height_std_dev = value.to_f
				when "evenness"
					scatter.evenness = value.to_f
				when "groups"
					#puts "processing groups"
					group_data = value
					group_data.each do |group_data|
						#puts "processing group"
						#puts "group_data: #{group_data.inspect}"
						
						group = ProtoObjectGroup.new
						
						group_data.each do |datum|
							#puts "datum: #{datum.inspect}"
							datum_val = datum[1]
							case datum[0]
							when "prototype_def_name"
								group.prototype_def_name = datum_val
							when "weight"
								group.weight = datum_val.to_f
							end
						end
						scatter.groups << group
					end
				end
			end
			
			scatter
		end
		
		#puts "@scatters: #{@scatters.inspect}"
	end
	
	
	def scatter_data(scatter)
		#puts "scatter_data(), scatter=#{scatter.inspect}"
		
		group_data = scatter.groups.map { |g| [ ["prototype_def_name", g.prototype_def_name], ["weight", g.weight] ] }
	
		s = scatter
		return [ 
			["groups", group_data], ["target_def_name", s.target_def_name], 
			["density", s.density], ["seed", s.seed], ["view_culling", s.view_culling], 
			["orient_normal", s.orient_normal], ["scale_std_dev", s.scale_std_dev], 
			["height_std_dev", s.height_std_dev], ["evenness", s.evenness]
		]
	end
	
	
	def save_scatter_settings
		#puts "save_scatter_settings()"
		
		scatters_data = @scatters.map { |s| scatter_data(s) }
		
		#puts "scatters_data: #{scatters_data.inspect}"
		
		@model.set_attribute(@dict, 'scatters', scatters_data)
		
		#read_back_attr = @model.get_attribute(@dict, "scatters", nil)
		#puts "read_back_attr: #{read_back_attr.inspect}"
	end

end
