#SkIndigo Ruby code is copyright Dale Martens 2009, all rights reserved.
require 'sketchup.rb'

#MDB Loaded List
#Ketchup
#Camera Rubber
#Brushed Aluminum
#Rippled Plastic
#Nitrile
#Scratched Rust
#Chocolate (null mat created)
#Red Candy Glass
#Gold
#Clay
#Milk Glass
#Portuguese Tiles
#Earth (had trouble unzipping due to file permissions)
#Etched Glass
#Peachy
#Checkerboard
#Beech Wood
#Brushed Metal
#Scratches
#Stone Surface
#White Ext Wall Plaster
#Galvanized Steel
#Floor Pinewood Plank
#Interior Wall Paint Roller
#Concrete Form
#Coors Beer
#Stone Ground
#Lampshade by BbB
#Tiled Linoleum
#Ghost Glass
#Parquet Wood Floor
#Elm Wood
#Brass Dull
#Tree Bark
#Cashew wood
#Clear Natural Water
#Ruby - (extra materials in IGM file)
#Old Parquet
#Fast Carpaint (contains extra materials)
#Simple Carpaint (contains extra materials)
#White Grapefruit Juice
#Apple Juice
#Ice
#Marble Floor Red
#Aluminum NK
#Simple Red Car Paint

#Working with version 2.4
#Diamond (reg tab)
#Sodium emission (reg tab)
#phillips ultralume flor. (reg tab)


#MDB Unloadable List
#Acier Corten  - Weathered Steel (shader textures not supported) - uv set 'uv' specified so could not link
#Yellow Brick (invalid PIGM?)
#Black Ash (wrong texture paths)
#Grass (wrong material name reference in blend material)
#Strawberry (displacement map in blend material) - uvset error
#Emerald (regular tabulated)
#Diamond fancy yellow (regular tabulated)
#Sapphire (regular tabulated)
#Diamond Fancy (regular tabulated)

#Brick and Grafitti - displacement map in blend material


#Bugs
#indigo bug - displacement_map tag allowed but not displacement tag in blend material

#Sketchup.send_action("showRubyPanel:")


class IndigoMaterialImporter

	def initialize(igm)

		@media=[]
		@materials=[]
		@num_materials=0
		@tex_path=File.dirname(igm)
		@mats=Sketchup.active_model.materials
		@null_mat=[]
		@irs=IndigoRenderSettings.new
		@verbose=false  #for debugging
		@new_materials=[]  #array of new IndigoMaterials created
	end #end init

	def replace_xml_reference(string)
		return string.gsub(/&lt;/,'<').gsub(/&gt;/,'>')
	end

	#####################returns an RGB triplet in the range 0-255 (backward compatibility)
	def get_bc_color(type,data)

		#TODO - check precedence of tags that set the color
		case type

		when "Diffuse","Oren-Nayar"
			colour_el=data["colour"]  #deprecated
			as=data["albedo_spectrum"]  #compatible with 1.0.9
			if colour_el
				colour=colour_el.element_text
				colour_gamma=2.2
			elsif as
				rgb=as["rgb"]
				if rgb
					colour=rgb["rgb"].element_text
					colour_gamma=rgb["gamma"].element_text
				end
			end

		when "Phong"
			colour=data["diffuse"] #deprecated as of version 0.9
			das=data["diffuse_albedo_spectrum"]   #compatible with 1.0.9
			srs=data["specular_reflectivity_spectrum"] #compatible with 1.0.9
			if colour
				colour=colour.element_text
				colour_gamma=2.2
			elsif das
				rgb=das["rgb"]
				if rgb
					colour=rgb["rgb"].element_text
					colour_gamma=rgb["gamma"].element_text
				end
			elsif srs
				rgb=srs["rgb"]
				if rgb
					colour=rgb["rgb"].element_text
					colour_gamma=rgb["gamma"].element_text
				end
			end

		end #end case

		if colour
			colour=SkIndigo.igm_rgb_gamma_convert(colour,colour_gamma.to_f,2.2)
			return colour
		else
			return nil
		end
	end  

	#####################sets the values of an Indigo Texture from an XML element
	def set_tex_param(im,itex,element)

		itex.a=element["a"].element_text if element["a"]
		itex.b=element["b"].element_text if element["b"]
		itex.c=element["c"].element_text if element["c"]
		itex.exponent=element["exponent"].element_text if element["exponent"]
		return itex
	end
	
	###############set a wavelength dependent parameter 
	def set_wd_param(im,itex,data,textures,set_sketchup,force_constant)  #indigo material, indigo texture,xml data, and array of xml textures
		#set_sketchup is a flag that will set the sketchup texture or color (for albedo or specular maps for example)
		#force constant is a flag for parameters that can not take shaders or textures (eg. g_spectrum, base emission)
		constant=data["constant"]
		texture=data["texture"]
		shader=data["shader"]

		if constant or force_constant
			if constant==nil
				constant=data  #this is a backwards compatibility thing with older versions of Indigo
			end
			itex.type="Constant"
			rgb=constant["rgb"]
			reg_tab=constant["regular_tabulated"]
			bb=constant["blackbody"]
			uni=constant["uniform"]
			if rgb and (set_sketchup==false)
				itex.spectrum_type="RGB"
				rgb_vals=rgb["rgb"]
				if rgb_vals
					rgb_vals=rgb_vals.element_text.split("\s")
					#p rgb_vals
					rgb_vals=rgb_vals.collect {|val| val.to_f}
					max_val=rgb_vals.max
					gain=1.0
					if max_val>1.0
						gain=max_val
						rgb_vals=rgb_vals.collect {|val| val/gain}
					end
					#p "rgb values"
					#p rgb_vals
					itex.rgb_gain=gain.to_s
					itex.rgb_r=rgb_vals[0].to_s
					itex.rgb_g=rgb_vals[1].to_s
					itex.rgb_b=rgb_vals[2].to_s
				end
				itex.rgb_gamma=rgb["gamma"].element_text if rgb["gamma"]
			elsif rgb and (set_sketchup==true)  #apply the settings to the sketchup color
				itex.type="SketchUp"
				colour=rgb["rgb"].element_text
				colour_gamma=rgb["gamma"].element_text
				colour=SkIndigo.igm_rgb_gamma_convert(colour,colour_gamma.to_f,2.2)
				red=(colour[0]*255).round
				green=(colour[1]*255).round
				blue=(colour[2]*255).round
				im.color=[red,green,blue]
			end
			
			if reg_tab
				itex.spectrum_type="Tabulated"
				itex.reg_num_values=reg_tab["num_values"].element_text if reg_tab["num_values"]
				itex.reg_start=(reg_tab["start_wavelength"].element_text.to_f*1E9).to_i if reg_tab["start_wavelength"]
				itex.reg_end=(reg_tab["end_wavelength"].element_text.to_f*1E9).to_i if reg_tab["end_wavelength"]
				itex.reg_values=reg_tab["values"].element_text if reg_tab["values"]
			end
			if bb
				itex.spectrum_type="Blackbody"
				itex.bb_temperature=bb["temperature"].element_text if bb["temperature"]
				itex.bb_gain=bb["gain"].element_text if bb["gain"]
			end
			if uni
				itex.spectrum_type="Uniform"
				itex.uni_value=uni["value"].element_text if uni["value"]
			end
	
		elsif texture
			tex_index=texture["texture_index"].element_text
			if tex_index
				this_tex=textures[tex_index.to_i]
				#p "\ntexture data found \n#{this_tex}"
				itex.type="External" if set_sketchup==false
				itex.type="SketchUp" if set_sketchup==true
				itex.a=this_tex["a"].element_text if this_tex["a"]
				itex.b=this_tex["b"].element_text if this_tex["b"]
				itex.c=this_tex["c"].element_text if this_tex["c"]
				itex.exponent=this_tex["exponent"].element_text if this_tex["exponent"]
				f_name=this_tex["path"].element_text
				tex_path=set_path(f_name)
				if FileTest.exist?(tex_path)
					itex.path=tex_path #if set_sketchup==false
					im.texture=tex_path if set_sketchup==true
					@irs.store_texture(tex_path)
					#p tex_path
					
					UI.messagebox "New External texture #{tex_path} found." if @verbose
				else
					UI.messagebox ("External texture #{tex_path} not found.")
				end
			end
		elsif shader
			itex.type="Shader"
			itex.shader_code=shader["shader"].element_text
		end

		return itex
	end
	
	###############set a wavelength independent parameter (bump, sigma, displace, exponent)
	def set_wi_param(im,itex,data,textures)  #indigo material, indigo texture,xml data, and array of xml textures

		constant=data["constant"]
		texture=data["texture"]
		shader=data["shader"]

		if constant
			itex.b=constant.element_text
			itex.type="Constant"
		elsif texture
			tex_index=texture["texture_index"].element_text
			if tex_index
				this_tex=textures[tex_index.to_i]
				#p "\ntexture data found \n#{this_tex}"
				itex.type="External"
				itex.a=this_tex["a"].element_text if this_tex["a"]
				itex.b=this_tex["b"].element_text if this_tex["b"]
				itex.c=this_tex["c"].element_text if this_tex["c"]
				itex.exponent=this_tex["exponent"].element_text if this_tex["exponent"]
				if this_tex["smooth"]
					itex.smooth="On" if this_tex["smooth"].element_text=="true"
					itex.smooth="Off" if this_tex["smooth"].element_text=="false"
				end
				f_name=this_tex["path"].element_text
				tex_path=set_path(f_name)
				if FileTest.exist?(tex_path)
					itex.path=tex_path
					@irs.store_texture(tex_path)
					UI.messagebox "New External texture #{tex_path} found." if @verbose
				else
					UI.messagebox ("External texture #{tex_path} not found.")
				end
			end
		elsif shader
			itex.type="Shader"
			itex.shader_code=shader["shader"].element_text
		else  #backward compatible - should work with old <exponent>500</exponent> tags
			itex.b=data.element_text.to_f   #validate data?
			itex.type="Constant"
		end

		return itex
	end


	###############sets blend map parameters
	def set_blend_param(im,itex,data,textures)  #indigo material, indigo texture,xml data, and array of xml textures

		constant=data["constant"]
		texture=data["texture"]
		shader=data["shader"]

		if constant
			itex.b=constant.element_text
			itex.type="Constant"
		elsif texture
			tex_index=texture["texture_index"].element_text
			if tex_index
				this_tex=textures[tex_index.to_i]   #get the xml data for this texture
				itex.type="SketchUp"
				itex.a=this_tex["a"].element_text if this_tex["a"]
				itex.b=this_tex["b"].element_text if this_tex["b"]
				itex.c=this_tex["c"].element_text if this_tex["c"]
				itex.exponent=this_tex["exponent"].element_text if this_tex["exponent"]
				f_name=this_tex["path"].element_text
				tex_path=set_path(f_name)
				if FileTest.exist?(tex_path)
					im.texture=tex_path
					itex.path=tex_path
					@irs.store_texture(tex_path)
					UI.messagebox "Material texture set to #{File.basename(tex_path)}" if @verbose
				else
					UI.messagebox "Blend texture #{tex_path} not found."
				end
				
			end
		elsif shader
			itex.type="Shader"
			itex.shader_code=shader["shader"].element_text
		end

		return itex #return the texture object
	end
	
	###############sets albedo map parameters
	def set_albedo_param(im,itex,data,textures)  #indigo material, indigo texture,xml data, and array of xml textures

		constant=data["constant"]
		texture=data["texture"]
		shader=data["shader"]

		if constant
			rgb=constant["rgb"] #check if rgb spectrum was specified for this attribute
			reg_tab=constant["regular_tabulated"]
			bb=constant["blackbody"]
			uni=constant["uniform"]
			if rgb
				itex.type="SketchUp"
				colour=rgb["rgb"].element_text
				colour_gamma=rgb["gamma"].element_text
				colour=SkIndigo.igm_rgb_gamma_convert(colour,colour_gamma.to_f,2.2)
				red=(colour[0]*255).round
				green=(colour[1]*255).round
				blue=(colour[2]*255).round
				im.color=[red,green,blue]
				UI.messagebox "SketchUp RGB values set to :R=#{red} G=#{green} B=#{blue}" if @verbose
			end
			if reg_tab
				itex.spectrum_type="Tabulated"
				itex.reg_num_values=reg_tab["num_values"].element_text if reg_tab["num_values"]
				itex.reg_start=reg_tab["start_wavelength"].element_text.to_f/1E9 if reg_tab["start_wavelength"]
				itex.reg_end=reg_tab["end_wavelength"].element_text.to_f/1E9 if reg_tab["end_wavelength"]
				itex.reg_values=reg_tab["values"].element_text.strip if reg_tab["values"]
			end
			if bb
				itex.spectrum_type="Blackbody"
				itex.bb_temperature=bb["temperature"].element_text if bb["temperature"]
				itex.bb_gain=bb["gain"].element_text if bb["gain"]
			end
			if uni
				itex.spectrum_type="Uniform"
				itex.uni_value=uni["value"].element_text if uni["value"]
			end
		elsif texture
			tex_index=texture["texture_index"].element_text
			if tex_index
				this_tex=textures[tex_index.to_i]   #get the xml element for this texture
				itex.type="SketchUp"
				itex.a=this_tex["a"].element_text if this_tex["a"]
				itex.b=this_tex["b"].element_text if this_tex["b"]
				itex.c=this_tex["c"].element_text if this_tex["c"]
				itex.exponent=this_tex["exponent"].element_text if this_tex["exponent"]
				f_name=this_tex["path"].element_text
				tex_path=set_path(f_name)
				if FileTest.exist?(tex_path)
					im.texture=tex_path
					UI.messagebox "Material texture set to #{File.basename(tex_path)}" if @verbose
				else
					UI.messagebox "Albedo texture #{tex_path} not found."
				end
				
			end
		elsif shader
			itex.type="Shader"
			itex.shader_code=shader["shader"].element_text
		end

		return itex #return the texture object
	end
	
	###############sets specular map parameters
	def set_spec_param(im,itex,data,textures)  #indigo material, indigo texture,xml element, and array of xml texture elements

		constant=data["constant"]
		texture=data["texture"]
		shader=data["shader"]

		if constant
			rgb=constant["rgb"] #check if rgb spectrum was specified for this attribute
			#uniform=parse("uniform",constant)  #check if uniform spectrum
			if rgb
				colour=rgb["rgb"].element_text
				colour_gamma=rgb["gamma"].element_text
				colour=SkIndigo.igm_rgb_gamma_convert(colour,colour_gamma.to_f,2.2)
				red=(colour[0]*255).round
				green=(colour[1]*255).round
				blue=(colour[2]*255).round
				im.color=[red,green,blue]
				UI.messagebox "SketchUp RGB values set to :R=#{red} G=#{green} B=#{blue}" if @verbose
			end
		elsif texture
			tex_index=texture["texture_index"].element_text
			if tex_index
				this_tex=textures[tex_index.to_i]   #get the xml element for this texture
				itex.type="SketchUp"
				itex.a=this_tex["a"].element_text if this_tex["a"]
				itex.b=this_tex["b"].element_text if this_tex["b"]
				itex.c=this_tex["c"].element_text if this_tex["c"]
				itex.exponent=this_tex["exponent"].element_text if this_tex["exponent"]
				f_name=this_tex["path"].element_text
				tex_path=set_path(f_name)
				if FileTest.exist?(tex_path)
					im.texture=tex_path
					UI.messagebox "Material texture set to #{File.basename(tex_path)}"
				else
					UI.messagebox "Specular Reflectivity texture #{tex_path} not found."
				end
				
			end
		elsif shader
			itex.type="Shader"
			itex.shader_code=shader["shader"].element_text
		end

		return itex #return the texture object

	end

	####set absorption and scattering parameters
	def set_media_param(im,data,medium_name)   #im is an indigo material, data is an xml element

		#######Extract and set opacity values#########
		type=im.type

		medium=@media.find {|element| element["name"].element_text==medium_name}
		#p medium["name"].element_text

		#medium is the medium element
		if medium
			im.precedence=medium["precedence"].element_text
			basic=medium["basic"]
			if basic
				im.ior=basic["ior"].element_text
				im.cauchy_b_coeff=basic["cauchy_b_coeff"].element_text
				acs=basic["absorption_coefficient_spectrum"]  #deprecated as of Indigo 2.4
				ac=basic["absorption_coefficient"]
				a_map=im.absorption_map
				if acs  #backward compatible routine
					a_map=set_wd_param(im,a_map,acs,nil,false,true)  #force constant
				elsif ac  #parameter could be constant or shader
					a_map=set_wd_param(im,a_map,ac,nil,false,false)  #sets the absorption map parameters
				end #if ac
				if a_map.spectrum_type=="RGB" #used for setting the SketchUp RGB and opacity
					a_map.type="SketchUp"
					acs_gamma=a_map.rgb_gamma
					acs_gain=a_map.rgb_gain.to_f
					rgb_text=[a_map.rgb_r,a_map.rgb_g,a_map.rgb_b].join(" ")
					rgb_text=[a_map.rgb_r.to_f*acs_gain,a_map.rgb_g.to_f*acs_gain,a_map.rgb_b.to_f*acs_gain].join(" ")
					p rgb_text
					#p "setting rgb values"
					acs_rgb_value=SkIndigo.igm_rgb_gamma_convert(rgb_text,acs_gamma.to_f,1.0)
				end	
				im.absorption_map=a_map  #stores the absorption parameters to the indigoMaterial
				
				sss=basic["subsurface_scattering"]
				if sss
					im.sss=true
					scs=sss["scattering_coefficient_spectrum"] #deprecated as of indigo 2.4
					sc=sss["scattering_coefficient"]
					sss_map=im.sss_map
					if scs
						sss_map=set_wd_param(im,sss_map,scs,nil,false,true) #force constant
					elsif sc  #parameter could be constant or shader
						sss_map=set_wd_param(im,sss_map,sc,nil,false,false)
					end
					im.sss_map=sss_map  #stores the parameters to the Indigo material
					
					pf=sss["phase_function"]
					hg=pf["henyey_greenstein"]
					if hg
						p "henyey-greenstein"
						im.phase_function="Henyey-Greenstein"
						gs=hg["g_spectrum"]
						gsm=im.g_spectrum_map
						im.g_spectrum_map=set_wd_param(im,gsm,gs,nil,false,true)
					else
						im.phase_function="Uniform"
					end
							
				end #end if sss
			end #if basic

		end #if medium

		###the following routine sets the SketchUp color and opacity based on the absorption properties of the medium.
		if (type=="Specular") or (type=="Glossy Transparent")		
		begin
			if acs_rgb_value
				rgb_ab=acs_rgb_value
				red_ab=rgb_ab[0]
				green_ab=rgb_ab[1]
				blue_ab=rgb_ab[2]
				opacity=(Math.sqrt([red_ab,green_ab,blue_ab].max*10)).round
				opacity=99 if opacity>99
				opacity=1 if opacity<1
				red=((1-(10*red_ab)/(opacity**2))*255).round
				red=0 if red<0
				red=255 if red>255
				blue=((1-(10*blue_ab)/(opacity**2))*255).round
				blue=0 if blue<0
				blue=255 if blue>255
				green=((1-(10*green_ab)/(opacity**2))*255).round
				green=0 if green<0
				green=255 if green>255
				set_alpha=opacity.to_f/100
				im.alpha=set_alpha
				im.color=[red,green,blue]
				UI.messagebox "Input Absorption Coefficient Spectrum (gamma=#{acs_gamma}):\nR=#{red_ab} G=#{green_ab} B=#{blue_ab}\nSketchUp opacity set to #{opacity}\nSketchUp RGB set to #{red} #{green} #{blue}\nExported Absorption Coefficient Spectrum (gamma=1.0):\nR=#{"%.4f" %((1-red.to_f/255)*opacity*opacity/10)} G=#{"%.4f" %((1-green.to_f/255)*opacity*opacity/10)} B=#{"%.4f" %((1-blue.to_f/255)*opacity*opacity/10)}" if @verbose
				
			#elsif acs_uniform_value
			#	
			#	opacity=Math.sqrt(acs_uniform_value.to_f*10).round
			#	opacity=99 if opacity>99
			#	opacity=1 if opacity<1
			#	new_color=((1-(10*acs_uniform_value.to_f)/(opacity**2))*255).round
			#	new_color=0 if new_color<0
			#	new_color=255 if new_color>255
			#	set_alpha=opacity.to_f/100
			#	im.alpha=set_alpha
			#	im.color=[new_color,new_color,new_color]
			#	UI.messagebox "Input Absorption Coefficient Spectrum:Uniform=#{acs_uniform_value}\nSketchUp opacity set to #{opacity}\nSketchUp RGB set to #{new_color} #{new_color} #{new_color}\nExported Absorption Coefficient Spectrum:R=#{"%.4f" %((1-new_color.to_f/255)*opacity*opacity/10)} G=#{"%.4f" %((1-new_color.to_f/255)*opacity*opacity/10)} B=#{"%.4f" %((1-new_color.to_f/255)*opacity*opacity/10)}"
			#	
			end		
			
		rescue
			UI.messagebox "Error setting opacity and RGB values.  Check IGM file for errors."
			raise
		end #end error handle

		end #end if transparent
	end



	############################XML Parsing Method
	def parse_old(tag,string)

		data=string.split("<"+tag+">")[1]
		data=string.split("<"+tag+">")[2] if tag=="rgb"  #this is used to parse the rgb values from an rgb spectrum element
		data=string.split("<"+tag+">")[2] if tag=="shader" #used to parse shader code from a shader element
		if data
			data=data.split("<\/"+tag+">")[0]
			return nil if not data
		else
			return nil
		end

		return data.strip
	end

	############################XML Parsing Method
	def parse(tag,string,depth=1)

		depth.times {|i|
			string="" if string==nil
			start_tag="<"+tag+">"
			end_tag="<\/"+tag+">"
			s_index=string.index(start_tag)
			if s_index  #found the start tag
				s_offset=start_tag.length+s_index
				r_index=string.rindex(end_tag)
				if r_index
					string=string[s_offset..r_index-1].strip
				else
					string=nil
				end
			else
				string=nil
			end
		}
		return string
	end


	########################XML parsing method
	def get_child(string)  #string is an xml string

		child_start_re=/<\w+>/
		#child_end_re=/<\/\w+>/

		pos=string=~child_start_re  #get the position of the first child tag found
		if pos  #if we found a tag
			tag_name=$&[1..-2]  #store child tag_name by collecting the contents of the match minus the <> symbols
			start_child_text=$'
			child_start_re=/<#{tag_name}>/
			child_end_re=/<\/#{tag_name}>/
			found_end_pos=start_child_text=~child_end_re  #try to find the end tag
			while found_end_pos!=nil
				#try to find another start tag in the child text
				found_start_pos=start_child_text=~child_start_re
				if found_start_pos<found_end_pos   #we have a nested tag
					start_child_text=start_child_text[found_end_pos]
				end
				found_end_pos=start_child_text
				#if we find any other start tags before the next matched end tag of this element, we ignore all data 
				found_start=nil
				found_start=start_child_text=~child_start_re  #try to find the start of a new xml tag
				if found_start
				
				#todo
				end
			end #end while
		end
	end


	##################################Function to acquire array of media from IGM file##################
	def get_media(data)

		media_arr=[]
		media=data.split("<medium>")
		if media.length>1 #we have at least one medium
			for i in (1..media.length-1)
				media_data=media[i].split("</medium>")[0]
				medium_name=parse("name",media_data) if media_data
				media_arr.push(medium_name) if medium_name
				media_arr.push(media_data) if media_data
			end
		else
			return nil
		end

		return media_arr
	end

	####################get an array of textures XML data (current as of version 1.1.x)
	def get_textures(data)

		tex_arr=[]
		textures=data.split("<texture>")
		if textures.length>1 #we have at least one texture
			for i in (1..textures.length-1)
				texture_data=textures[i].split("</texture>")[0]
				if texture_data  #found a <texture> tag, however, we also found texture indices tags
					unless texture_data["<texture_index>"]  #throw out the texture indices tags, we are not interested
						tex_arr.push(texture_data.strip)
					end
				end
			end
		else
			return nil
		end

		#p tex_arr
		return tex_arr
	end
	
	
	##################################Function to acquire array of materials from IGM file##################
	def get_materials(data)

		material_arr=[]
		materials=data.split("<material>")

		if materials.length>1 #we have at least one material
			for i in (1..materials.length-1)
				material_data=materials[i].split("</material>")[0]
				material_arr.push(material_data) if material_data
			end
		else
			return nil
		end

		return material_arr
	end


	# Read a boolean element from some XML element.
	# For example:
	# read_bool_elem(coating, "interference", false)
	#
	def read_bool_param(mat_elem, bool_name, default_value)
		bool_elem = mat_elem[bool_name]
		if bool_elem
			value_str = bool_elem.element_text
			return true  if value_str=="true"
			return false if value_str=="false"
			
			raise "Boolean element '#{bool_name}' had invalid value: '#{value}'"
		else
			return default_value
		end
	end


	##########################
	def load_igm(igm,mat)

		@mat=mat
		model=Sketchup.active_model
		model.start_operation "Load IGM"

		begin
		@mat.reset()   #reset the material and delete all old indigo attributes
		#open and read the igm file
		igm_file=File.new(igm,"r")
		md=igm_file.read
		igm_file.close

		md = SkIndigo.remove_invalid_utf8_sequences(md)

		xml_parser=DMxml::Parser.new(md)
		doc=xml_parser.parse_document()

		#get the <materialdata> for compatibility
		material_data=doc["materialdata"]
		material_data=doc["scenedata"] if not material_data
		if material_data==nil
			UI.messagebox("Warning: This does not appear to be a valid IGM file.")
		end

		unsupported=["epidermis","dermis","peak","private_data","param"]
		unsupported.each {|att|
			found=doc.find_element(att)
			if found   #found an unsupported indigo attribute
				result=UI.messagebox("This material contains an Indigo attribute <#{found.name}> that cannot be represented in SkIndigo. This material cannot be loaded by SkIndigo but will render correctly if you create a Linked IGM.  Would you like to create a Linked IGM material instead?", MB_YESNO)
				if result==6
					return 1
				else
					return nil
				end
			end
		}

		@media=material_data.find_all("medium")
		@materials=material_data.find_all("material") #returns array of all materials in the IGM file

		for material in @materials
			
			colour=nil
			textures=[]
			@num_materials+=1
			#p "material #{@num_materials}"
			
			material_name=material["name"].element_text #get the name element of the material
			material_name=replace_xml_reference(material_name)
			
			#First check if the material is a null material
			if material["null_material"] or material["null_material "]
				@null_mat.push(material_name)
				#p "Found null material #{material_name}"
				next
			end
			
			unless (@materials.length)==@num_materials  #create a new SU material for each material found except when this is the last material in the list
				im=IndigoMaterial.new(@mats.add(material_name.gsub(/\*/,"~")))
				@new_materials.push(im)
				UI.messagebox "New material created named #{material_name}." if @verbose
			else
				im=@mat  #if this is the last material in the IGM file, we will load the properties into the currently selected material
			end
			
			#load and set the material type
			specular=material["specular"] #found a specular material
			if specular
				im.type="Specular"
				textures=specular.find_all("texture")
				transparent=specular["transparent"].element_text
				im.alpha=1.0 if transparent=="false"
				im.alpha=0.5 if transparent=="true"  #ensure that the SU opacity is less than 1.0
				imn=specular["internal_medium_name"].element_text
				imn=replace_xml_reference(imn)
				set_media_param(im,specular,imn)
				load_common_mat_params(im,specular,textures)
				load_common_media_mat_params(im,specular,textures)
				load_bc_textures(im,specular)
			end
			
			glossy_transparent=material["glossy_transparent"]  #found a glossy transparent material
			if glossy_transparent
				#p "glossy transparent"
				im.type="Glossy Transparent"
				textures=glossy_transparent.find_all("texture")
				imn=glossy_transparent["internal_medium_name"].element_text
				imn=replace_xml_reference(imn)
				set_media_param(im,glossy_transparent,imn)
				load_common_mat_params(im,glossy_transparent,textures)
				load_exponent_param(im,glossy_transparent,textures)
				load_common_mat_params(im,glossy_transparent,textures)
				load_common_media_mat_params(im,glossy_transparent,textures)
				load_bc_textures(im,glossy_transparent)
			end
			
			diffuse=material["diffuse"] #found a diffuse material
			if diffuse
				#p "diffuse"
				im.type="Diffuse"
				textures=diffuse.find_all("texture")
				colour=get_bc_color("Diffuse",diffuse)  #backward compatibility color check
				im.alpha=1.0
				load_albedo_param(im,diffuse,textures)
				load_common_mat_params(im,diffuse,textures)
				load_bc_textures(im,diffuse)
			end
			
			diffuse_transmitter=material["diffuse_transmitter"]  #found a diffuse transmiter material
			if diffuse_transmitter
				#p "diffuse transmitter"
				im.type="Diffuse"
				colour=get_bc_color("Diffuse",diffuse_transmitter)
				textures=diffuse_transmitter.find_all("texture")
				im.alpha=0.0  #in SkIndigo, a pure diffuse transmitter material is defined by a diffuse material with opacity=0.0
				load_albedo_param(im,diffuse_transmitter,textures)
				load_common_mat_params(im,diffuse_transmitter,textures)
				load_bc_textures(im,diffuse_transmitter)
			end
			
			phong=material["phong"]  #found a phong material
			if phong
				#p "phong"
				im.type="Phong"
				colour=get_bc_color("Phong",phong)
				textures=phong.find_all("texture")
				im.alpha=1.0
				nk_data=phong["nk_data"]
				if nk_data and nk_data.element_text.length>0
					im.type="Metals"
					im.use_specular=false
					######current IGM spec is to only use nkdata located in root indigo directory
					#path=File.dirname(SkIndigo.get_indigo_path)  unless SkIndigo.on_mac? #the root indigo directory
					#path=SkIndigo.get_indigo_app_data_path if SkIndigo.on_mac?
					nkpath=File.join(nk_data.element_text.gsub(/\\/,"\/"))
					#nkpath=File.join(path,nk_data.element_text.gsub(/\\/,"\/"))   #this may have to change depending on the final IGM spec
					nk=NKdata.new(nkpath)
					#check if the nkdata path is valid
					if (not nk.valid?)
						UI.messagebox "Could not find nkdata file #{File.basename(nkpath)} in indigo nkdata folder.  Please manually load nkdata."
					else
						im.nk_data=nk    #set the nkdata path
						im.use_nkdata=true   #set the flag to use the nkdata
					end
				end
				
				#if this tag is present, it overrides any diffuse albedo tags
				srs=phong["specular_reflectivity_spectrum"]
				sr=phong["specular_reflectivity"]
				if srs or sr
					sr=srs if sr==nil
					im.type="Metals"
					im.use_nkdata=false
					#im.specular_map=set_spec_param(im,im.specular_map,sr,textures)
					im.specular_map=set_wd_param(im,im.specular_map,sr,textures,true,false)
				end
				
				if im.type=="Phong"   #these attributes are not necessary to parse if the specular reflectivity element or nkdata is present (Metals)
					im.ior=phong["ior"].element_text if phong["ior"]
					#im.fresnel_scale=parse("fresnel_scale",phong) #deprecated
					load_albedo_param(im,phong,textures)
				end
				load_exponent_param(im,phong,textures)
				load_common_mat_params(im,phong,textures)
				load_bc_textures(im,phong) 
			end
			
			blend=material["blend"]  #found a blend material
			if blend
				#p "blend"
				im.type="Blend"
				im.alpha=1.0
				textures=blend.find_all("texture")
				#Set any null materials found in a blend material to none
				a_name=blend["a_name"].element_text
				a_name=replace_xml_reference(a_name)
				if a_name
					#p "a_name #{a_name}"
					im.a_name=a_name.gsub(/\*/,"~")
					im.a_name="None" if @null_mat.index(a_name)
				end
				b_name=blend["b_name"].element_text
				b_name=replace_xml_reference(b_name)
				if b_name
					#p "b_name #{b_name}"
					im.b_name=b_name.gsub(/\*/,"~")
					im.b_name="None" if @null_mat.index(b_name)
				end
				step_blend=blend["step_blend"]
				if step_blend
					step_blend=step_blend.element_text
					im.step_blend=true if step_blend=="true"
					im.step_blend=false if step_blend=="false"
				end
				load_bc_textures(im,blend)
				blend2=blend["blend"]  #current in 1.1.X
				if blend2
					im.blend_map=set_blend_param(im,im.blend_map,blend2,textures)
				end
				unless im.blend_map?
					blend_factor=blend["blend_factor"] #used in 1.0.9
					if blend_factor
						bm=im.blend_map
						bm.b=blend_factor.element_text
						bm.type="Constant"  #set to constant
						im.blend_map=bm
					end
				end
				#convert a material blended with null to a clip mapped material
				load_clip_map(im) if im==@mat  #only load clip map for the parent material
				
			end #end if blend
			
			oren_nayar=material["oren_nayar"]   #found an oren-nayar material
			if oren_nayar
				#p "oren-nayar"
				im.type="Oren-Nayar"
				im.alpha=1.0
				textures=oren_nayar.find_all("texture")
				colour=get_bc_color("Oren-Nayar",oren_nayar)
				load_albedo_param(im,oren_nayar,textures)
				sigma=oren_nayar["sigma"]
				im.sigma_map=set_wi_param(im,im.sigma_map,sigma,textures) if sigma
				load_common_mat_params(im,oren_nayar,textures)
				load_bc_textures(im,oren_nayar)
			end
			
			coating=material["coating"]   #found a coating material
			if coating
				# puts "coating"
				im.type="Coating"
				im.alpha=1.0
				textures=coating.find_all("texture")
				
				# Load absorption
				absorption = coating["absorption"]
				im.absorption_map = set_wd_param(im, im.absorption_map, absorption, textures, false,false) if absorption
				
				# Load thickness
				thickness = coating["thickness"]
				im.thickness_map = set_wi_param(im, im.thickness_map, thickness, textures) if thickness
				
				# Load roughness
				roughness = coating["roughness"]
				im.roughness_map = set_wi_param(im, im.roughness_map, roughness, textures) if roughness
				
				# Load fresnel scale  (Called fresnel_map in SkIndigo)
				fresnel_scale = coating["fresnel_scale"]
				im.fresnel_map = set_wi_param(im, im.fresnel_map, fresnel_scale, textures) if fresnel_scale
				
				# Load IOR
				im.ior=coating["ior"].element_text if coating["ior"]
				
				# Load interference
				im.interference = read_bool_param(coating, "interference", false)

				# Load substrate material
				substrate_material_name = coating["substrate_material_name"].element_text
				substrate_material_name = replace_xml_reference(substrate_material_name)
				if substrate_material_name
					# puts "substrate_material_name: #{substrate_material_name}"
					im.substrate = substrate_material_name.gsub(/\*/,"~")
				end
				
				load_common_mat_params(im,coating,textures)
			end

			double_sided_thin=material["double_sided_thin"]   #found a double-sided thin material
			if double_sided_thin
				# puts "double_sided_thin"
				im.type="Double-Sided Thin"
				im.alpha=1.0
				textures=double_sided_thin.find_all("texture")
				
				# Load r_f
				r_f = double_sided_thin["r_f"]
				im.r_f_map = set_wi_param(im, im.r_f_map, r_f, textures) if r_f
				
				# Load transmittance
				transmittance = double_sided_thin["transmittance"]
				im.transmittance_map = set_wd_param(im, im.transmittance_map, transmittance, textures, false,false) if transmittance
				

				# Load front roughness
				front_roughness = double_sided_thin["front_roughness"]
				im.front_roughness_map = set_wi_param(im, im.front_roughness_map, front_roughness, textures) if front_roughness
				
				# Load back roughness
				back_roughness = double_sided_thin["back_roughness"]
				im.back_roughness_map = set_wi_param(im, im.back_roughness_map, back_roughness, textures) if back_roughness
				
						
				# Load front fresnel scale (front_fresnel_scale_map)
				front_fresnel_scale = double_sided_thin["front_fresnel_scale"]
				im.front_fresnel_scale_map = set_wi_param(im, im.front_fresnel_scale_map, front_fresnel_scale, textures) if front_fresnel_scale
				
				# Load back fresnel scale (back_fresnel_scale_map)
				back_fresnel_scale = double_sided_thin["back_fresnel_scale"]
				im.back_fresnel_scale_map = set_wi_param(im, im.back_fresnel_scale_map, back_fresnel_scale, textures) if back_fresnel_scale
				
				# Load IOR
				im.ior=double_sided_thin["ior"].element_text if double_sided_thin["ior"]
				
				
				# Load front material
				front_material_name = double_sided_thin["front_material_name"].element_text
				front_material_name = replace_xml_reference(front_material_name)
				if front_material_name
					# puts "front_material_name: #{front_material_name}"
					#im.front_material_name = front_material_name.gsub(/\*/,"~")
					im.set("front_material_name", front_material_name.gsub(/\*/,"~"))
				end
				
				# Load back material
				back_material_name = double_sided_thin["back_material_name"].element_text
				back_material_name = replace_xml_reference(back_material_name)
				if back_material_name
					# puts "back_material_name: #{back_material_name}"
					#im.back_material_name = back_material_name.gsub(/\*/,"~")
					im.set("back_material_name", back_material_name.gsub(/\*/,"~"))
				end
				
				load_common_mat_params(im,double_sided_thin,textures)
			end
					
			#######Extract and set RGB Values#############Backwards Compatible
			if (colour) and (im.texture==nil) 		
				red=(colour[0]*255).round
				green=(colour[1]*255).round
				blue=(colour[2]*255).round
				im.color=[red,green,blue]
				UI.messagebox "SketchUp RGB values set to :R=#{red} G=#{green} B=#{blue}" if @verbose
					
			end #end if colour
		end #end for mat in materials

		model.commit_operation

		rescue
			UI.messagebox("An error occured while loading this IGM file.")
			model.abort_operation
			raise

		end #begin

		return 0

	end #end load igm

	#####
	def load_common_mat_params(im,element,textures)  #indigo material, material xml element (i.e. <diffuse>)

		bump=element["bump"]
		im.bump_map=set_wi_param(im,im.bump_map,bump,textures) if bump
		im.use_bump=true if bump
		normal=element["normal_map"]
		im.normal_map=set_wi_param(im,im.normal_map,normal,textures) if normal
		im.use_normal=true if normal
		displace=element["displacement"]
		im.disp_map=set_wi_param(im,im.disp_map,displace,textures) if displace
		im.use_displace=true if displace
		base_emission=element["base_emission"]
		im.base_emission=set_wd_param(im,im.base_emission,base_emission,nil,false,true) if base_emission
		im.layer=0 if base_emission
		emission=element["emission"]
		im.emission_map=set_wd_param(im,im.emission_map,emission,textures,false,false) if emission
		im.use_emission=true if emission
	end #load common mat

	####
	def load_common_media_mat_params(im,element,textures)

		abs_layer_trans=element["absorption_layer_transmittance"]
		im.abs_layer_trans_map=set_wd_param(im,im.abs_layer_trans_map,abs_layer_trans,textures,false,false) if abs_layer_trans
		im.use_abs_layer_trans=true if abs_layer_trans				
	end

	####
	def load_exponent_param(im,element,textures)

		exponent=element["exponent"]
		im.exponent_map=set_wi_param(im,im.exponent_map,exponent,textures) if exponent
	end

	###
	def load_albedo_param(im,element,textures)

		albedo=element["albedo"]
		albedo=element["diffuse_albedo"] if albedo==nil  #for phong materials
		im.albedo_texture=set_wd_param(im,im.albedo_texture,albedo,textures,true,false) if albedo
	end

	###
	def load_clip_map(im)

		a_name=im.a_name
		b_name=im.b_name
		blend_map=im.blend_map
		clip_map=IndigoTexture.new(blend_map.to_s)  #effecitvely copies the blend map paramters to the clip map
		if im.blend_map?  #todo extend to constant blend params
			if a_name=="None" and b_name!="None"
				clip_map.type="External"
				#clip_map.path=File.join(@tex_path,im.texture.filename)
				im2=IndigoMaterial.find(b_name)  #get the other material
				im2.copy(im)  #copy all the attributes from one to the other
				im.clip_map=clip_map.invert
				im.use_clip=true
				im2.delete!  #this material is no longer required
			end
			if b_name=="None" and a_name!="None"
				clip_map.type="External"
				#clip_map.path=File.join(@tex_path,im.texture.filename)
				im2=IndigoMaterial.find(a_name)  #get the other material
				im2.copy(im)  #copy all the attributes from one to the other
				im.clip_map=clip_map
				im.use_clip=true
				im2.delete!  #this material is no longer required
			end
		end
	end

	########backward compatible texture loading
	def load_bc_textures(im,elem)

		#Backwards Compatibility Methods
		#version 1.0.9 albedo texture laoding
		albedo_tex=elem["albedo_texture"]
		if albedo_tex
			albedo_path=albedo_tex["path"].element_text
			if FileTest.exist?(set_path(albedo_path))
				im.texture=albedo_path
				UI.messagebox "Material texture set to #{File.basename(albedo_path)}" if @verbose
			else
				UI.messagebox "Albedo texture #{albedo_path} not found."
			end
			ab=im.albedo_texture
			ab=set_tex_param(im,ab,albedo_tex)
			ab.type="SketchUp"
			im.albedo_texture=ab
		end
		
		
		#version 1.0.9 bump map loading
		bump_map=elem["bump_map"]
		if bump_map
			im.use_bump=true
			bump_map_path=bump_map["path"].element_text
			# old methods for internal textures
			#tex_name=IndigoMaterial.find_texture_file(File.join(tex_path,bump_map_path))
			#su_mat_name=tex_name.display_name if tex_name
			bump_map_path=set_path(bump_map_path)
			bm=im.bump_map
			if FileTest.exist?(bump_map_path)
				bm.type="External"
				@irs.store_texture(bump_map_path)
				bm.path=bump_map_path
				UI.messagebox("Bump map set to #{bump_map_path}") if @verbose
			else
				UI.messagebox "Bump map #{bump_map_path} not found."
			end
			bm=set_tex_param(im,bm,bump_map)
			im.bump_map=bm
		end
		
		#version 1.0.9 exponent map loading
		exp_map=elem["exponent_map"]
		if exp_map
			exp_map_path=exp_map["path"].element_text
			#tex_name=IndigoMaterial.find_texture_file(File.join(tex_path,exp_map_path))
			#em.su_mat_name=tex_name.display_name if tex_name
			exp_map_path=set_path(exp_map_path)
			em=im.exponent_map
			if FileTest.exist?(exp_map_path)
				em.type="External"
				em.path=exp_map_path
				@irs.store_texture(exp_map_path)
				UI.messagebox "Exponent map set to #{exp_map_path}" if @verbose
			else
				UI.messagebox "Exponent map #{exp_map_path} not found."
			end
			em=set_tex_param(im,em,exp_map)
			im.exponent_map=em
		end
		
		#version 1.0.9 displacement map loading
		disp_map=elem["displacement_map"]
		if disp_map
			im.use_displace=true
			disp_map_path=disp_map["path"].element_text
			#tex_name=IndigoMaterial.find_texture_file(File.join(tex_path,disp_map_path))
			#dm.su_mat_name=tex_name.display_name if tex_name
			disp_map_path=set_path(disp_map_path)
			dm=im.disp_map
			if FileTest.exist?(disp_map_path)
				dm.type="External"
				dm.path=disp_map_path
				@irs.store_texture(disp_map_path)
				UI.messagebox "Displacement map set to #{disp_map_path}" if @verbose
			else
				UI.messagebox "Displacement map #{disp_map_path} not found."
			end
			dm=set_tex_param(im,dm,disp_map)
			im.disp_map=dm
		end
		
		#version 1.0.9 blend map loading
		blend_map=elem["blend_map"]
		if blend_map
			bm=im.blend_map
			blend_map_path=blend_map["path"].element_text
			blend_map_path=set_path(blend_map_path)
			if FileTest.exist?(blend_map_path)
				im.texture=blend_map_path
				bm.path=blend_map_path
				@irs.store_texture(blend_map_path)
				UI.messagebox "Material texture set to #{File.basename(blend_map_path)}" if @verbose
			else
				UI.messagebox "Blend map #{blend_map_path} not found."
			end
			bm=set_tex_param(im,bm,blend_map)
			bm.type="SketchUp"
			im.blend_map=bm
		end

	end  #end backward compatible texture loading

	###can take either a relative or absolute path - will return an absolute path
	def set_path(path)

		if path[1]==58 #absolute path found on PC
			return path
		end
		if path[0]==47 #absolute path found on Mac
			return path
		end
		return File.join(@tex_path,path)  #if relative path, join to path of igm file
	end
		
	end #end class