#SkIndigo Ruby code is copyright Dale Martens 2011, all rights reserved, and
#Glare Technologies Limited, 2014 - 
#-----------------------------------------------------------------------------
# Name        	: SkIndigo.rb
# Description 	: Model exporter and material editor for Indigo Renderer
# Menu Item   	: Plugins\SkIndigo\
# Author      	: Dale Martens (Whaat), Glare Technologies Limited


#require 'profiler'
def reinterpret_as_utf8(str)
	if String.method_defined?(:encode)
		str = str.encode('UTF-16', 'UTF-8', :invalid => :replace, :replace => '') # Convert to UTF-16, assuming encoding is UTF-8
		str = str.encode('UTF-8', 'UTF-16') # Convert back to UTF-8
	end
	return str
end


def convert_to_utf8(str)
	if String.method_defined?(:encode)
		return str.encode('UTF-8')
	end
	return str
end


skindigo_path=File.dirname(reinterpret_as_utf8(__FILE__))
load File.join(skindigo_path,"SkIndigoUVS.rb")
load File.join(skindigo_path,"ObjectEditor.rb")
load File.join(skindigo_path,"IndigoMeshBuilder.rb")
load File.join(skindigo_path,"IndigoRenderSettings.rb")
load File.join(skindigo_path,"IndigoSettingsEditor.rb")
load File.join(skindigo_path,"DMXML.rb")
load File.join(skindigo_path,"indigoMatImport.rb")
load File.join(skindigo_path,"AnimationExport.rb")
load File.join(skindigo_path,"sun_override.rb")
load File.join(skindigo_path,"material_picker.rb")
load File.join(skindigo_path,"DMcolormath.rb")
load File.join(skindigo_path,"IndigoSpectrumPicker.rb")
load File.join(skindigo_path,"SkindigoMatObserver.rb")
load File.join(skindigo_path,"IndigoExporter.rb")
load File.join(skindigo_path,"IndigoMesh.rb")
load File.join(skindigo_path,"IndigoTexture.rb")
load File.join(skindigo_path,"IndigoMaterial.rb")
load File.join(skindigo_path,"IndigoScattersEditor.rb")




HAVE_BYTESIZE = "hello".respond_to?("bytesize")


module SkIndigo

	#
	# In SketchUp 2014 this class was changed from Set to Sketchup::Set
	# in order to avoid conflict with the Ruby Standard Library. 
	# from: http://www.sketchup.com/intl/en/developer/docs/ourdoc/set
	if defined?(Sketchup::Set)
		Set = Sketchup::Set
	end


	######test if user is on a mac platform
	def SkIndigo.on_mac?
		return (Object::RUBY_PLATFORM =~ /darwin/i) ? true : false
	end
	
	def SkIndigo.ruby_is_version_2_or_later
		return RUBY_VERSION.split(".")[0].to_i >= 2
	end
end


$:.push(File.join(File.dirname(reinterpret_as_utf8(__FILE__))))  #add the skindigo folder to the ruby library search list
require 'sketchup.rb'

unless SkIndigo.on_mac?
	require 'Win32API'
	require 'win32ole'
	require 'registry'
	
	# Net/http is only in SketchUp 2014 or later which corresponds to ruby version 2.0.0 or later
	if SkIndigo.ruby_is_version_2_or_later
		require 'net/http'
		require 'uri'
	end
end

module SkIndigo

	def SkIndigo.verbose?()
		false
	end
	
	@@run_tests = nil # Cache if ENV["SKINDIGO_RUN_TESTS"] is true
	def SkIndigo.run_tests?
		if @@run_tests.nil?
			@@run_tests = ENV["SKINDIGO_RUN_TESTS"] == "TRUE"
		end
		return @@run_tests
	end
	
	def SkIndigo.test_render_time
		return 0.2
	end
	
	# Should be 8, 13, 14 etc..
	# SketchUp 2014 has major version 14.
	def SkIndigo.sketchup_major_version
		Sketchup.version.split('.')[0].to_i
	end
	
	def SkIndigo.sketchup_is_2014_or_later
		return sketchup_major_version >= 14
	end
	

	def SkIndigo.start_operation(name)

		if Sketchup.version.split(".")[0].to_i>=7
			Sketchup.active_model.start_operation(name,true)
		else
			Sketchup.active_model.start_operation(name)
		end
	end

	def SkIndigo.commit_operation
		Sketchup.active_model.commit_operation
	end

	def SkIndigo.validate_export_selected_entities_only
		return Sketchup.active_model.selection.length > 0 ? MF_ENABLED : MF_GRAYED
	end

	def SkIndigo.default_export_path()
		return SkIndigo.get_user_app_support_path
	end
	
	#####export the scene with a user-given filename
	def SkIndigo.export_as(mode, textures)  

		model=Sketchup.active_model
		model_filename = File.basename(model.path,".skp")
		if model_filename!=""
			model_name = model_filename
			model_name += ".igs"
		else
			model_name = "Untitled.igs" 	
		end

		def_path = File.join(model_name)	# default path for exporting is the indigo directory
		
		# Unfortunately, as of SketchUp 2015, it seems to be impossible to set both the suggest filename and the file type ('Save as Type') filter at the same time in the Windows save dialog.
		# The whole thing seems rather buggy.
		# See bug SU-0165 on SketchUp bug tracker.
		export_path = UI.savepanel("Export Model", "" , def_path)
		
			
		return if export_path==nil
		export_path = File.join(File.dirname(export_path), File.basename(export_path, ".*") + ".igs")
		SkIndigo.export(:normal, true, export_path, false, false)
	end

	###### Export and Render ###########
	# mode should be one of :normal, or :mat_preview.
	# Returns true if export was successful, or false if it failed, due to e.g. an exception being thrown and an error dialog displayed.
	def SkIndigo.export(mode, export_textures, export_path, auto_render, should_export_selected_entities_only)
		# Export Modes
		# :normal = Normal (exporting meshes)
		# :mat_preview = Material Preview Scene

		puts "SkIndigo.export(): mode: #{mode}" if SkIndigo.verbose?
		
		begin
			Sketchup.set_status_text("Exporting...")	
				
			is = IndigoRenderSettings.new
			indigo_path = SkIndigo.get_indigo_path
			if indigo_path == nil
				UI.messagebox "Indigo application not found.  Export aborted."
				return false
			end
			if !export_path  
				last_path = is.last_path  # Returns the last path that the model was exported to
				if (last_path == nil) or (!FileTest.exist?(last_path))
					exp_path = SkIndigo.default_export_path()
					export_path = File.join(exp_path, "default.igs")
				else
					export_path=last_path
				end
			end
			ie = IndigoExporter.new(export_path)
			
			ie.preview_scene = true if mode==:mat_preview
			
			model = ie.model
			model_name = File.basename(ie.path, ".igs")
			return false if model_name == ""
			tex_path = File.join(File.dirname(export_path), "TX_" + model_name)
			ie.tex_path = "TX_" + model_name
			
			renderingoptions = model.rendering_options
			render_mode = renderingoptions["RenderMode"]
			
			# TODO fix and test monochrome/clay mode	
			# monochrome mode currently disabled
			if render_mode == 5 # Tests for monochrome render mode
				ie.clay_render = true	
			end
			
			print_timings = !SkIndigo.run_tests?
		
			ie.start = Time.new
			mesh_builder = IndigoMeshBuilder.new
			scene_element = "scene"
			if mode == :normal			#normal export mode (exporting meshes)
				# Work out if we want to export only the selected entities, or all entities in scene.
				if should_export_selected_entities_only && Sketchup.active_model.selection.length>0
					entities_to_export = Sketchup.active_model.selection
				else
					entities_to_export = Sketchup.active_model.entities
				end
				
				IndigoMeshBuilder.reset  # Erases the stored meshes, instances, and spheres from last export

				# Process entities (iterate over scene, get list of components/groups etc.. to export)
				t = Time.now
				Sketchup.set_status_text("Collecting Instances...")
				puts "SkIndigo.export(): calling mesh_builder.process_entities()..." if SkIndigo.verbose?
				mesh_builder.process_entities(entities_to_export)
				puts "Traversed model and got instances in #{Time.now-t} seconds." if print_timings
				
				# Build required (visible, referenced) meshes
				t = Time.now
				Sketchup.set_status_text("Building Meshes...")
				puts "SkIndigo.export(): calling mesh_builder.build_meshes()..." if SkIndigo.verbose?
				mesh_builder.build_meshes()
				puts "Built meshes in #{Time.now-t} seconds." if print_timings

				ie.get_used_layers()
			
			elsif mode == :mat_preview #preview scene
				# TODO: fix
				ents = Sketchup.active_model.entities
				IndigoMeshBuilder.reset 		
				#mesh_builder.build_active_mesh(ents, )
				mesh_builder.get_instances(ents, Geom::Transformation.new, exported_entity_set)
				mesh_builder.build_meshes()
				ie.get_used_layers()
				scene_element = "scenedata"
			end	
			
			# Open the file to export to.
			out = SkIndigo.new_writable_utf8_file(export_path)
			out.puts "<?xml version=\"1.0\" encoding=\"utf-8\"?>"
			out.puts "<#{scene_element}>"
			
			ie.export_metadata(out)
			ie.export_render_settings(out)
			ie.export_section_planes(out)
			ie.export_tonemapping(out)
			ie.export_camera(model.active_view, out)
			ie.export_environment(out)
			ie.export_default_mat(out)
			
			
			# Export SketchUp fog
			ie.export_fog(out) if model.rendering_options["DisplayFog"]
			
			# Build a set of the materials that were actually referenced by a model, so that we can just export those materials, and not all materials in the scene.
			referenced_materials = ie.get_referenced_materials(mesh_builder)
			
			#----------- Export all the materials -------------
			# Don't export the previewmaterial if we are creating a previewscene
			if mode == :mat_preview
				referenced_materials = referenced_materials.delete_if {|m, generate_uvs| m.name == "previewmaterial"}
			end
			
			puts "SkIndigo.export(): calling export_material()..." if SkIndigo.verbose?
			t = Time.now
			Sketchup.set_status_text("Exporting Materials...")
			referenced_materials.each {|mat, generate_uvs|
				ie.export_material(IndigoMaterial.new(mat), out, generate_uvs)
			}
			puts "Exported materials in #{Time.now-t} seconds."	if print_timings
			#------------------------------------------------
			
			
			# Image export 
			t = Time.now
			Sketchup.set_status_text("Exporting Images...")
			mesh_builder.images.each {|image| ie.export_image(image,out)}
			puts "Exported images in #{Time.now-t} seconds." if print_timings
			
			ie.export_light_layer_names(out)
			
			# Write meshes
			t = Time.now
			Sketchup.set_status_text("Exporting Meshes...")
			ie.export_meshes(mesh_builder, out)
			puts "Exported meshes in #{Time.now-t} seconds." if print_timings
			
			# Write instances
			t = Time.now
			Sketchup.set_status_text("Exporting Instances...")
			mesh_builder.spheres.each {|sphere| ie.export_sphere(sphere, out)}
			
			puts "SkIndigo.export(): calling export_instances()..." if SkIndigo.verbose?
			ie.export_instances(mesh_builder, out)
			puts "Exported instances in #{Time.now-t} seconds." if print_timings
			
			
			# Write scatters
			ie.export_scatters(mesh_builder, out)
				
			
			out.puts "</#{scene_element}>"

			#ie.validate_light_source()
			
			out.close
			
			# If exporting is successful, store the IGS path as an attribute.  Don't store it if it hasn't change, to avoid showing a 'save changes?' prompt.
			if mode != :mat_preview && is.last_path != ie.path
				is.last_path = ie.path
			end
			
			#------------------Export textures---------------
			if export_textures
				t = Time.now
				Sketchup.set_status_text("Exporting Textures...")
				
				Dir.mkdir(tex_path) if !FileTest.exist?(tex_path)

				# Export all textures collected during the material export process
				ie.export_textures(ie.textures)

				# Export all collected ies profiles
				for path in ie.ies
					profile = IESProfile.new(path)
					profile.export(tex_path) if profile.valid?  # Export the IES profiles to the textures directory
				end
				puts "Exported textures in #{Time.now-t} seconds." if print_timings
			end
			
			#------------------End export textures---------------
			
			
			if print_timings
				puts "------SkIndigo Export Run------"	
				puts "Instances: #{mesh_builder.instances.size}"
				puts "Meshes: #{mesh_builder.num_meshes()}"
				puts "Meshes exported: #{mesh_builder.num_new_meshes_exported}"
				puts "Materials exported: #{referenced_materials.length}"
				puts "Spheres: #{mesh_builder.spheres.length}"
			end
			
			Sketchup.set_status_text("Export Complete.")	
			ie.end = Time.new
			elapsed = ie.end-ie.start
			puts "Exported in #{"%0.2f" % elapsed} seconds."
			
			return true if mode == :mat_preview # Exit here if exporting a preview scene
			
			$skindigo_exported_this_session = true
			
			if !auto_render
				result = UI.messagebox("Model saved in file:\n" + export_path + "\nWould you like to render now?", MB_YESNO) 
				auto_render = true if result == 6
			end
			
			if ie.export_warnings.size > 0
				UI.messagebox("There were some problems while exporting: \n" + ie.export_warnings.join("\n"))
			end	
			
			#--------------Open Indigo for rendering
			if auto_render && FileTest.exist?(indigo_path)
				indigo_root = File.dirname(indigo_path)
				renders_dir = SkIndigo.get_renders_folder()

				#check if user wants to use a custom image name
				output_param = ""
				case is.network_mode
				when 1 # master
					output_param += "-n m " unless SkIndigo.on_mac?
				when 2 # working master
					output_param += "-n wm " unless SkIndigo.on_mac?
				end
				if is.custom_image_name?
					image_name = is.image_name
					output_path = File.join(renders_dir, image_name + ".png")
					igi_output_path = File.join(renders_dir, image_name + ".igi")
					output_param += "-o \"#{output_path}\" -igio \"#{igi_output_path}\"" unless SkIndigo.on_mac?
				end
				
				SkIndigo.launch_indigo(indigo_path,ie.path,output_param,is.low_priority?,"normal")		
			end
			return true
		rescue => exception
			UI.messagebox(
				"An error occurred while SkIndigo was exporting the scene.\n" + 
				"Please copy and paste the following details into an email to support@indigorenderer.com:\n" + 
				"\n" + 
				"SkIndigo version: " + VERSION + "\n" + 
				"SketchUp version: " + Sketchup.version.to_s + "\n" + 
				"Ruby platform: " + Object::RUBY_PLATFORM + "\n" + 
				"Ruby version: " + RUBY_VERSION + "\n" +
				"Error during processing: #{$!} \n" + 
				exception.backtrace.join("\n\t") + "\n",
				MB_MULTILINE, 
				"Error occurred in SkIndigo"
			) 
			return false
		end # End exception rescue
	end

	
	#####get the renders directory
	def SkIndigo.get_renders_folder()

		app_data = SkIndigo.get_indigo_app_data_path()
		renders = File.join(app_data, "renders")
		Dir.mkdir(renders) unless FileTest.exist?(renders)
		return renders
	end
	
	
	def SkIndigo.download_file(url, local_path)
	
		if SkIndigo.on_mac?
			command_line = "curl -o " << local_path << " " << url
			system(command_line)
		else
			# Net/http is only in SketchUp 2014 or later which corresponds to ruby version 2.0.0 or later
			if SkIndigo.sketchup_is_2014_or_later
				uri = URI(url)
		
				Net::HTTP.start(uri.hostname) do |http|
					resp = http.get(uri.path)
					open(local_path, "wb") do |file|
						file.write(resp.body)
					end
				end
			else
				#puts "Downloading with wget..."
				wget = File.join(File.dirname(reinterpret_as_utf8(__FILE__)), "wget.exe")
				Dir.chdir(File.dirname(wget))
				command_line = "start wget.exe -O " << local_path << " " << url
				#puts command_line
				system(command_line)
				time = 0.0
				while win_process_running?("wget.exe")
					sleep(1)
					time += 1
					if time > 60
						break
					end
				end
			end
		end
	end
	
	
	#########
	def SkIndigo.win_process_running?(p_name)
		wmi = WIN32OLE.connect("winmgmts://")
		query = "select * from Win32_process"
		processes = wmi.ExecQuery(query)
		names=[]
		processes.each{|p| names.push(p.Name)}
		return true if names.include?(p_name)
	end
	
	
	# SkIndigo.run_indigo_console_blocking("-v")
	def SkIndigo.run_indigo_console_blocking(arg_string)
		#puts "run_indigo_console_blocking()"
		indigo_path = SkIndigo.get_indigo_path
		
		console_path = File.join(File.dirname(indigo_path), "indigo_console.exe")
	
		exec_process_blocking(console_path, arg_string)
	end
	
	
	def SkIndigo.exec_process_blocking(process_path, arg_string)
		#puts "exec_process_blocking()"
	
		#puts "process_path: #{process_path}"
		
		process_filename = File.basename(process_path)
		
		#puts "process_filename: #{process_filename}"
		
		#first, store the pids of all the instances of indigo_console that are running
		wmi = WIN32OLE.connect("winmgmts://")
		query = "select * from Win32_process where name='#{process_filename}'"
		processes = wmi.ExecQuery(query)
		pids=[]
		processes.each { |process| pids.push(process.ProcessID) }
		
		#puts "initial pids: #{pids.inspect}"
				
		# Launch process
		window_param = 0 # hidden
		shell = WIN32OLE.new('Shell.Application')
		shell.ShellExecute(process_path, arg_string, '', 'open', window_param)
				
		#check if a new process of indigo_console was created and compare the pid to any previous ones
		wmi = WIN32OLE.connect("winmgmts://")
		query = "select * from Win32_process where name='#{process_filename}'"
		processes = wmi.ExecQuery(query)
		new_pids=[]
		processes.each { |process| new_pids.push(process.ProcessID) }
		
		#puts "new_pids pids: #{new_pids.inspect}"
		
		new_pid = (new_pids-pids)[0]  #return the new pid of the new instance of indigo (or nil if a new indigo_console process was not found)
		
		#puts "new_pid: #{new_pid.inspect}"
		
		if new_pid.nil?
			# d'oh, just wait for a while
		else
			# Wait for process to finish.
			#puts "waiting for process..."
			while true
				wmi = WIN32OLE.connect("winmgmts://")
				query = "select * from Win32_process where name='#{process_filename}'"
				processes = wmi.ExecQuery(query)
				pids=[]
				processes.each { |process| pids.push(process.ProcessID) }
				
				if pids.include?(new_pid) # If process is still running...
					#puts "sleeping..."
					sleep(0.05)
				else
					break
				end
			end
		end

		#puts "done."
	end


	########method to launch indigo application with optional parameters
	def SkIndigo.launch_indigo(indigo_path,igs_file_path,params="",low_priority=false,window="max")

		if SkIndigo.on_mac?  #we can currently only open the indigo app on a mac (cannot launch scene)
			#UI.openURL("file:///"+File.expand_path(indigo_path))
			#indigo_path=File.expand_path(indigo_path)
			#p indigo_path
			Dir.chdir(File.dirname(indigo_path))
			indigo_exe=File.basename(indigo_path)
			#export_path="\"#{File.expand_path(igs_file_path)}\""
			export_path="#{File.expand_path(igs_file_path)}"
			#/Applications/Indigo.app/Contents/MacOS/Indigo path_to_file
			if indigo_exe.downcase=~/indigoconsole/
				indigo_exe=File.expand_path(indigo_path)
				command_line="'"<<indigo_exe<<"' '"<<export_path<<"' "<<params
				#p command_line
				#Process.kill(0,preview_thread+1)
				#Process.wait
				system("kill #{$skindigo_preview_thread}") if $skindigo_preview_thread
				pid=fork {exec(command_line)}
				return pid
				#system("killall IndigoConsole")
				#Process.kill("KILL",$skindigo_preview_thread) if $skindigo_preview_thread
				#$skindigo_preview_thread=Thread.new {system(command_line)}
			else
				#command_line="open -a "<<indigo_exe<<" "<<export_path<<" "<<params
				#command_line
				#p "launching indigo"
				#system(command_line) #execute the command
				indigo_exe=File.join(File.expand_path(indigo_path),"Contents","MacOS","Indigo")
				command_line="'"<<indigo_exe<<"' '"<<export_path<<"' "<<params
				#p command_line
				#Process.kill(0,preview_thread+1)
				#Process.wait
				#system("kill #{$skindigo_preview_thread}") if $skindigo_preview_thread
				pid=fork {exec(command_line)}
			end

		else   #on windows, we can launch with scene and parameters
			# TODO - test this on windows - taskkill /F /IM <processname.exe> /T
			case window 
				when "max"
					window_param=3
				when "min"
					window_param=2
				when "normal"
					window_param=1
				when "hidden"
					window_param=0
			end
			Dir.chdir(File.dirname(indigo_path))
			indigo_exe=File.basename(indigo_path)
			low=""
			if low_priority
				low="\/low " 
				params=params+" --thread_priority belownormal"
			end
			export_path="\"#{igs_file_path}\""
			
			if SkIndigo.run_tests?
				params += " -halt #{SkIndigo.test_render_time} --settings \"<verbose>true</verbose><close_gui_on_halt>true</close_gui_on_halt>\""
			end

			begin
				############
				#launch_indigo_with_batch_file(indigo_path,igs_file_path,params,low_priority,window)
				#first, store the pids of all the instances of indigo_console that are running
				wmi = WIN32OLE.connect("winmgmts://")
				query = "select * from Win32_process where name='indigo_console.exe'"
				processes = wmi.ExecQuery(query)
				pids=[]
				processes.each {|process|
					pids.push(process.ProcessID)
				}
				
				#now create a new indigo process
				#puts indigo_exe
				#puts export_path
				#puts params
				shell = WIN32OLE.new('Shell.Application')
				shell.ShellExecute(indigo_exe, export_path+" "+params,'', 'open', window_param)
				
				#check if a new process of indigo_console was created and compare the pid to any previous ones
				wmi = WIN32OLE.connect("winmgmts://")
				query = "select * from Win32_process where name='indigo_console.exe'"
				processes = wmi.ExecQuery(query)
				new_pids=[]
				processes.each {|process|
					new_pids.push(process.ProcessID)
				}
				return (new_pids-pids)[0]  #return the new pid of the new instance of indigo (or nil if a new indigo_console process was not found)
			rescue Exception => msg
				puts "An exception occurred: #{msg}"
			
				# Launch Indigo.  if this was requested to be a hidden window, e.g. for material previews, use the --noninteractive flag.
				# In either case, we want to use indigo.exe, not indigo_console.exe, to avoid display of the console window.
				indigo_exe = "indigo.exe" if indigo_exe == "indigo_console.exe"
				
				noninteractive_arg = window == "hidden" ? " --noninteractive" : ""
				cmd = indigo_exe + " " + export_path + " " + params + noninteractive_arg
				# puts "cmd: '#{cmd}'"
				process = IO.popen(cmd)
				return process.pid
			end
			#i_proc=DMprocess.new
			#i_proc.create_process(indigo_exe+" "+export_path+" "+params)
		end
	end


	##########Display the render settings dialog - TODO - make the settings editor a singleton class and revise this method
	def SkIndigo.render_settings
		if not $g_skindigo_settings
			$g_skindigo_settings = IndigoSettingsEditor.new
		end
		$g_skindigo_settings.show
	end
	
	
	def SkIndigo.show_scatter_editor
		if not $g_skindigo_scatter_editor
			$g_skindigo_scatter_editor = IndigoScattersEditor.new
		end
		$g_skindigo_scatter_editor.show
	end


	#####About message
	def SkIndigo.about
		UI.messagebox("SkIndigo " + VERSION + "\n" + 
		"SketchUp Exporter for Indigo " + INDIGO_VERSION_REQUIRED + " and later.\n" +
		"Authors: Whaat, and Glare Technologies Limited\n\n" + 
		"Please visit Indigo's website and forums if you have questions about Indigo.\n" + 
		"http://www.indigorenderer.com", MB_MULTILINE , "About SkIndigo") 
	end

	#######method to list all materials and Indigo material types
	def SkIndigo.mat_query

		table=""
		model = Sketchup.active_model
		materials = model.materials

		materials.each do |m| 
					
			im=IndigoMaterial.new(m)
		  
			table+=m.name+" - "
			if im.emitter?
				table+=im.type.upcase+" (Emitter)\n"
			else
				table+=im.type.upcase+"\n"
			end
			next
		end
		  
		# Now display the results
		
		#header="MATERIAL NAME".ljust(40)
		#header+="TYPE\n"
		#UI.messagebox table,MB_OK,"Indigo Material Types"
		UI.messagebox table,MB_MULTILINE,"Indigo Material Types"
	end

	
	# Open a file in write mode, with UTF-8 encoding.
	def SkIndigo.new_writable_utf8_file(path)
		# The ':UTF-8' bit only works in SketchUp 2014 or later, as it requires Ruby 1.9.x or later.
		if SkIndigo.sketchup_is_2014_or_later
			return File.new(path, "w:UTF-8")
		else
			return File.new(path, "w")
		end
	end
	

	##############get the user's temp directory (platform dependent)
	@@cached_temp_dir = nil
	def SkIndigo.temp_dir
		if on_mac?
			return ENV['TMPDIR']
		else
			#t = Time.now
			#SkIndigo.correct_path(ENV['TEMP'])
			#puts "SkIndigo.correct_path took #{Time.now - t} s"
			
			begin
				# Try and convert path to correct unicode path
				@@cached_temp_dir = SkIndigo.correct_unicode_path(ENV['TEMP']) if !@@cached_temp_dir
				return @@cached_temp_dir
			rescue
				puts "path correction failed: #{$!}"
				return ENV['TEMP']
			end
		end
	end


	#######gets the indigo path from the windows registry
	def SkIndigo.get_indigo_path_from_registry

		begin
			Win32::Registry::HKEY_CURRENT_USER.open('Software\Glare Technologies\Indigo Renderer') do |reg|
				reg_typ, reg_val = reg.read('InstallDirectory')
				return reg_val
			end
		rescue
			raise
			#return nil
		end
	end

	#######gets the indigo version from the windows registry
	def SkIndigo.get_indigo_version_from_registry

		begin
			Win32::Registry::HKEY_CURRENT_USER.open('Software\Glare Technologies\Indigo Renderer') do |reg|
				reg_typ, reg_val = reg.read('InstalledVersion')
				return reg_val
			end
		rescue
			return nil
		end
	end


	def SkIndigo.check_indigo_version

		indigo_version=SkIndigo.get_indigo_version
		reqd=SkIndigo::INDIGO_VERSION_REQUIRED
		if reqd
			parsed_version=reqd.split(".")
			v1=parsed_version[0].to_i  #deletes the v character if it exists and converts to an integer
			v2=parsed_version[1].to_i
			v3=parsed_version[2].to_i
			reqd=[v1,v2,v3]
		end

		return if indigo_version==nil

		valid=true
		if indigo_version[0]<reqd[0]
			valid=false
		end
		#UI.messagebox(valid)
		if ((indigo_version[0]==reqd[0]) and (indigo_version[1]<reqd[1]))
			valid=false
		end
		#UI.messagebox(valid)
		if ((indigo_version[0]==reqd[0]) and (indigo_version[1]==reqd[1]) and (indigo_version[2]<reqd[2]))
			valid=false
		end

		#UI.messagebox(valid)
		if valid==false
			UI.messagebox("Your Indigo Version is #{indigo_version.join(".")}. You need version #{reqd.join(".")} or greater in order to use this version of SkIndigo.  Please upgrade your copy of Indigo or downgrade your SkIndigo plugin to version #{indigo_version.join(".")} or lower.")
			UI.openURL("http://www.indigorenderer.com")
		end
	end
	
	
	
	######
	def SkIndigo.get_indigo_version

		if SkIndigo.on_mac?
			data=SkIndigo.get_indigo_version_mac
		else
			data=SkIndigo.get_indigo_version_pc
		end

		if data!=nil
			reg=/(\d+)(\.)(\d+)(\.)(\d+)/
			version=data.slice(reg)
			#UI.messagebox(version)
			if version
				parsed_version=version.split(".")
				v1=parsed_version[0].to_i  
				v2=parsed_version[1].to_i
				v3=parsed_version[2].to_i
				return [v1,v2,v3]
			else
				p "Could not parse version"
			end
		end
	end
	
	
	def SkIndigo.get_indigo_version_mac

		begin
			indigo_path=SkIndigo.get_indigo_console_path
			Dir.chdir(File.dirname(indigo_path))
			indigo_exe=File.expand_path(indigo_path)
			return nil unless (indigo_exe=~/onsole/) 
			cmd="\"#{indigo_exe}\" -v"
			data=`#{cmd}`
			return data	
		rescue
			p "Could not get Indigo version on Mac"
			return nil
		end
	end
	
	
		# Convert to ASCII, replacing any Unicode characters with '?'.
	def SkIndigo.convert_to_ascii(s)
		return s.encode("ASCII", {:undef=>:replace, :replace=>'?'})
	end
	
	def SkIndigo.has_unicode_chars(s)
		return !(s.bytes.to_a.find { |b| b >= 128 }.nil?)
	end
	
	# Ruby garbles UTF-8 strings in several different ways - 
	# it may return the actual bytes, but thinks the encoding is ascii (does this for Euro sign. Use reinterpret_as_utf8() in this case)
	# or it may have replaced the unicode characters with '?'.  Use convert_to_ascii() for this case.
	# Ruby may also just return some seemingly random bytes, in which case we can't match by name.
	def SkIndigo.filenames_match(utf8name, b)
		return (convert_to_ascii(utf8name) == convert_to_ascii(b)) || 
			(utf8name == reinterpret_as_utf8(b))
	end
	
	def SkIndigo.dir_is_writable(dir_path)
		# Try and write a file to see if this dir is owned by the current user
		begin
			filepath = File.join(dir_path, "skindigo_writable_test.txt")
			
			testfile = File.new(filepath, "w")
			testfile.close
			
			# Remove the test file, don't worry about possible failure.
			begin
				File.delete(filepath)
			rescue
			end
			
			return true
		rescue
			return false
		end
	end
	
	# Due to ruby bugs, it's very difficult to read proper Unicode paths, from env vars etc..
	# Dir.entries(path, {:encoding => "UTF-8"}) does however work correctly though, returning proper UTF-8 file names.
	# So use it to reconstruct the paths in UTF-8.
	# SkIndigo.correct_unicode_path(ENV['APPDATA'])
	# SkIndigo.correct_unicode_path(ENV['TEMP'])
	def SkIndigo.correct_unicode_path(p)	
		# This method relies on using Dir.entries with {:encoding => "UTF-8"}, which won't work in the older Rubys in SketchUp 8 and 2013.
		return p if !SkIndigo.sketchup_is_2014_or_later
		
		verbose = false
	
		# First step: convert all backslashes to forwards slashes
		p = p.gsub(/\\/, '/')

		# Split into path components
		components = p.split('/')
		
		# This will be the resulting new path
		new_components = [components[0]]

		for i in 1...components.size
			puts "-------looping, i: #{i}----------\n\n" if verbose
					
			# Enumerate entries in the current dir, getting the results in UTF-8.  Need a trailing slash here to work correctly.
			entries = Dir.entries(new_components.join('/') + '/', {:encoding => "UTF-8"})
			
			puts "entries: #{entries.inspect}\n\n" if verbose
			
			#puts "components[i]: #{components[i]}"
			#puts "convert_to_ascii(components[i]): #{convert_to_ascii(components[i])}"
			#puts "convert_to_ascii(components[i]).bytes.to_a: #{convert_to_ascii(components[i]).bytes.to_a}"
			
			num_matches = entries.count { |e| SkIndigo.filenames_match(e, components[i]) }
			puts "num_matches: #{num_matches}" if verbose
			
			match = nil
			if num_matches == 0
				# If there are no dirs that match by name, pick a unicode dir that we can write to.
				match = entries.find { |e| has_unicode_chars(e) && dir_is_writable(new_components.join('/') + "/" + e) }
				if !match
					raise "Failed to correct path '#{p}', could not match directory '#{components[i]}' to one on disk."
				end
				
			elsif num_matches == 1 # If exactly one matched directory, use it.
				match = entries.find { |e| SkIndigo.filenames_match(e, components[i]) }
			else
				# Else if there are two or more matched directories, take the one that is writeable by the current user.
				match = entries.find { |e| SkIndigo.filenames_match(e, components[i]) && dir_is_writable(new_components.join('/') + "/" + e) }
			end
			
			new_components << match
			
			puts "new_components: #{new_components.inspect}\n\n" if verbose
			#puts "new_components.join('/').encoding: #{new_components.join('/').encoding}\n\n"
		end
		
		return new_components.join('/')
	end
	



	def SkIndigo.get_indigo_version_mac

		begin
			indigo_path=SkIndigo.get_indigo_console_path
			Dir.chdir(File.dirname(indigo_path))
			indigo_exe=File.expand_path(indigo_path)
			return nil unless (indigo_exe=~/onsole/) 
			cmd="\"#{indigo_exe}\" -v"
			data=`#{cmd}`
			return data	
		rescue
			p "Could not get Indigo version on Mac"
			return nil
		end
	end
	
	
	#######gets the indigo version
	def SkIndigo.get_indigo_version_pc

		begin
			indigo_exe=SkIndigo.get_indigo_console_path  #this method will return the Indigo GUI path if the console path was not found
			return nil unless (indigo_exe=~/console/)  
			log_path=File.join(SkIndigo.temp_dir,"indigo_version.txt")  #standard output will be directed to this file
			bf_path=File.join(SkIndigo.temp_dir,"get_indigo_version.bat")  #batch file that will launch 'indigo_console.exe -v'. 
			
			if FileTest.exist?(log_path)
				File.delete(log_path)
			end
			
			#create the batch file
			bf=File.new(bf_path,"w")
			bf.puts "\"#{indigo_exe}\" -v > \"#{log_path}\""  #output is directed to the log_path file
			bf.close()
			
			#execute the batch file - using ShellExecute is the only way I have found to launch indigo without opening a window
			window_param=0 #hidden window
			params=""
			shell = WIN32OLE.new('Shell.Application')
			shell.ShellExecute(bf_path, params,'', 'open', window_param)
			
			#now we have to give ShellExecute enough time to create and write to the file
			tries=0
			while (!FileTest.exist?(log_path))
				sleep(0.25)
				tries+=1
				return nil if tries==4  #something must have failed if it takes this long to create the file
			end

			#now read the file that should contain the output from indigo_console.exe -v
			f=File.new(log_path,"r")
			data=f.gets
			f.close()
			return data
		rescue
			p "Errors getting Indigo version from PC"
			return nil
		end
	end


	########method for users to set the indigo path
	def SkIndigo.set_indigo_path

		on_mac = SkIndigo.on_mac?
		if on_mac
			path = SkIndigo.get_user_app_support_path
			path = File.join(path, "indigo_path.txt")
		else
			path = File.join(SkIndigo.get_indigo_app_data_path, "indigo_path.txt")
		end

		indigo_path = UI.openpanel("Locate Indigo.exe","","*.exe") if not on_mac
		indigo_path = UI.openpanel("Locate Indigo Application","","*.exe") if on_mac
		if indigo_path
			path_file=File.new(path,"w") 
			path_file.write(indigo_path) 
			path_file.close 
		else
			UI.messagebox("In order to render with Indigo, you must download and install the Indigo Renderer application.")
		end

		if (FileTest.exist?(indigo_path)) and (File.basename(indigo_path,".*").downcase=="indigo") #check if the path to indigo is valid
			#p indigo_path
			return indigo_path
		else
			return nil
		end
	end

	
	#####
	def SkIndigo.get_user_app_support_path

		on_mac = SkIndigo.on_mac?
		if on_mac
			path = File.join(ENV['HOME'], "Library", "Application Support", "SkIndigo")
			Dir.mkdir(path) unless FileTest.exist?(path)
		else
			path = SkIndigo.get_indigo_app_data_path
		end
		return path
	end

	#####
	def SkIndigo.get_indigo_path_from_program_files(app_name)

		program_files=ENV['ProgramFiles'] 
		indigo_dir=File.join(program_files,app_name) if program_files
		if (program_files) and FileTest.exist?(File.join(indigo_dir,"indigo.exe"))
			return File.join(indigo_dir,"indigo.exe")
		else
				
			program_files=ENV['ProgramFiles(x86)'] 
			indigo_dir=File.join(program_files,app_name) if program_files
			if (program_files) and FileTest.exist?(File.join(indigo_dir,"indigo.exe"))
				return File.join(indigo_dir,"indigo.exe")
			else
				program_files=ENV['ProgramW6432'] 
				indigo_dir=File.join(program_files,app_name) if program_files
				if (program_files) and FileTest.exist?(File.join(indigo_dir,"indigo.exe"))
					return File.join(indigo_dir,"indigo.exe")
				end
			end
		end
		return nil
	end


	# Try and find the path to the Indigo Renderer or Indigo RT install by looking in some common locations on disk.
	# Returns the full path to the Indigo executable if found, e.g. "C:\Program Files (x86)\Indigo Renderer\indigo.exe" or "/Applications/Indigo.app"
	# If no Indigo executable can be found, returns nil.
	def SkIndigo.detect_indigo_path
		# puts "detect_indigo_path()"
		if !SkIndigo.on_mac?
			path = SkIndigo.get_indigo_path_from_program_files("Indigo Renderer")
			path_rt = SkIndigo.get_indigo_path_from_program_files("Indigo RT")
			if path != nil
				# puts "detect_indigo_path(): Found path #{path}"
				return path
			elsif path_rt != nil
				return path_rt
			end
			
		else   #user is on Mac OSX
			
			# Try Indigo Renderer
			path = "/Applications/Indigo Renderer/Indigo.app"
			return path if FileTest.exist?(path)
			
			# Try Indigo RT
			path = "/Applications/Indigo RT/Indigo.app"
			return path if FileTest.exist?(path)
		end
		
		return nil
	end


	####################method to get path to indigo.exe or indigo.app (mac)
	def SkIndigo.get_indigo_path

		begin 
			path = File.join(SkIndigo.get_user_app_support_path, "indigo_path.txt")

			if FileTest.exist?(path) # Check if indigo_path.txt exists
				path_file = File.new(path, "r")
				indigo_path = path_file.read # Get the contents of indigo_path.txt file
				path_file.close
					
				if FileTest.exist?(indigo_path) and (File.basename(indigo_path,".*").downcase == "indigo") # Check if the path to indigo is valid
					#p indigo_path
					return indigo_path
				else
					indigo_path = SkIndigo.set_indigo_path # Ask user to set the path
				end
			else
				indigo_path = SkIndigo.detect_indigo_path() # Try and detect the path by looking in known locations on disk.
				if indigo_path == nil
					indigo_path = SkIndigo.set_indigo_path # Ask user to set the path
				end
			end

			if FileTest.exist?(indigo_path) and (File.basename(indigo_path,".*").downcase=="indigo") # Check if the path to indigo is valid
				#p indigo_path
				return indigo_path
			else
				return nil
			end
		rescue
		end

		UI.messagebox("SkIndigo is having trouble finding the Indigo Renderer Application.  If you are on a PC, please install Indigo Renderer in your default Program Files directory.  On a Mac, please install Indigo in the Applications folder.")
	end
	
	
	##############################################################
	def SkIndigo.get_indigo_console_path

		path=SkIndigo.get_indigo_path
		return nil if not path

		if SkIndigo.on_mac?
			c_path=File.join(path,"Contents","MacOS","IndigoConsole")
		else
			root=File.dirname(path)
			c_path=File.join(root,"indigo_console.exe")
		end
		if FileTest.exist?(c_path)
			return c_path
		else
			return path  #return the indigo GUI path if the console path is not found
		end
	end

	
	#########
	# Returns the 'resource path'.  On Windows this is the directory indigo.exe is in.
	# On OS X this will be Contents/Resources in the app bundle directory.
	def SkIndigo.get_resource_dir_path
		if SkIndigo.on_mac?
			indigo_app_dir = SkIndigo.get_indigo_path
			return File.join(indigo_app_dir, "Contents", "Resources")
		else
			return File.dirname(SkIndigo.get_indigo_path)
		end
	end
	
	
	#########
	@@cached_app_data_path = nil
	def SkIndigo.get_indigo_app_data_path

		if SkIndigo.on_mac?
			#indigo_app=SkIndigo.get_indigo_path
			#app_data=File.join(indigo_app,"Contents","Resources")
			app_data = File.join(ENV['HOME'], "Library", "Application Support", "Indigo Renderer")
			Dir.mkdir(app_data) unless FileTest.exist?(app_data)
			return app_data
		else
			# Ruby messes this up, See https://bugs.ruby-lang.org/issues/5570
			begin
				@@cached_app_data_path = SkIndigo.correct_unicode_path(ENV['APPDATA']) if !@@cached_app_data_path
				app_data = @@cached_app_data_path
			rescue
				puts "path correction failed: #{$!}"
				app_data = ENV['APPDATA'] # Fall back to using env var directly.
			end

			indigo_app_data = File.join(app_data, "Indigo Renderer")
			unless FileTest.exist?(indigo_app_data)
				Dir.mkdir(indigo_app_data)
			end
			return indigo_app_data
		end
	end
	
	
	
	#######method to pack an igm
	def SkIndigo.pack_igm(igm_path)

		return nil if !FileTest.exist?(igm_path)
		#p igm_path
		icp=SkIndigo.get_indigo_console_path
		return nil if not icp
		root=File.dirname(icp)

		#create the output path for the PIGM file
		out=File.basename(igm_path,".igm")
		out+=".pigm"
		out_dir=File.dirname(igm_path)
		out=File.join(out_dir,out)

		Dir.chdir(root) #changes the working directory to the indigo directoy
		cmd="#{icp} --pack \"#{igm_path}\" \"#{out}\""
		status=system(cmd)  #executes the command
		#p status
		return nil if not status
		UI.messagebox("IGM packed to #{out}")
		return out
	end

	
	
	#############################################################
	def SkIndigo.unpack_pigm(pigm)   #returns the path of the igm that was unpacked

		begin

			return nil if !FileTest.exist?(pigm)
			icp=SkIndigo.get_indigo_console_path
			return nil if not icp

			root=File.dirname(icp)
			igm_name=File.basename(pigm,".pigm")
			igm=igm_name+".igm"

			unless SkIndigo.on_mac?
				unpacked_igm_path = File.join(SkIndigo.get_indigo_app_data_path, "unpacked")
				unless FileTest.exist?(unpacked_igm_path)
					Dir.mkdir(unpacked_igm_path)
				end
				out=File.join(unpacked_igm_path,igm_name)
				unless FileTest.exist?(out)
					Dir.mkdir(out)
				end
				Dir.chdir(root)  #changes the working directory to the indigo directoy
				#cmd="indigo_console.exe --unpack \"#{pigm}\" \"#{out}\""
				#status=system(cmd)
				
				cmd = "--unpack \"#{pigm}\" \"#{out}\""
				#puts "cmd: #{cmd}"
				SkIndigo.run_indigo_console_blocking(cmd)
				
				unpacked_pigm_dir=out
			else
				indigo_console=File.expand_path(icp)
				unpacked_igm_path="\/tmp"
				unless FileTest.exist?(unpacked_igm_path)
					Dir.mkdir(unpacked_igm_path)
				end
				system("cp \"#{pigm}\" \/tmp")   #copies the PIGM file to temp dir
				Dir.chdir("\/tmp")
				#p Dir.pwd
				system("mkdir \"#{igm_name}\"")	
				cmd="\""<<indigo_console<<"\" --unpack \"#{File.basename(pigm)}\" \"#{igm_name}\""   #unpack the PIGM file to the temp directory
				#p cmd
				status=system(cmd)
				user_app_data=SkIndigo.get_user_app_support_path
				unpacked_pigm_dir=File.join(user_app_data,igm_name)
				unless FileTest.exist?(unpacked_pigm_dir)
					Dir.mkdir(unpacked_pigm_dir)
				end
				cmd="mv \"#{igm_name}\" \"#{user_app_data}\""
				system(cmd)
				
				raise if not status
			end

			

			#try to find the IGM file in the unpacked PIGM
			igm_dir=Dir.entries(unpacked_pigm_dir)
			igm=igm_dir.find {|f| f.split(".")[-1]=="igm"}
			return nil if igm==nil  #could not find an igm file in the unpacked file
			igm_path=File.join(unpacked_pigm_dir,igm)
			#p "igm path is #{igm_path}"
			return igm_path
		rescue
			UI.messagebox("Errors occurred while attempting to unpack the PIGM file.")
			raise
		end
	end


	######
	def SkIndigo.set_layer_names()

		irs=IndigoRenderSettings.new
		layer_names=irs.light_layer_names
		prompts=layer_names.collect {|layer| layer_names.index(layer).to_s}
		values=layer_names
		results=UI.inputbox(prompts,values,"Light Layer Names")
		return if not results
		irs.light_layer_names=results
		$skindigo_mat_observer.refresh_dialog()
		if $g_skindigo_settings
			$g_skindigo_settings.refresh_dialog()
		end
	end

	
	
	##################dummy instancing function#############
	def SkIndigo.dummy(object)

		if object.class==Sketchup::Group
			return object.entities if object.name==""
			arr=object.name.split("_")
		else
			arr=object.definition.name.split("_")
		end
		if (arr.length>1) and (arr.last.upcase=="DUMMY")
			arr.pop
			dummy_def_name=arr.join("_")
			if dummy_def_name
				def_list=Sketchup.active_model.definitions
				dummy_def=def_list[dummy_def_name]
			end
		end
		if dummy_def
			return dummy_def.entities
		else
			if object.class==Sketchup::Group
				return object.entities
			else
				return object.definition.entities
			end
		end
	end #function end

	
	
	##########################
	def SkIndigo.purge_mats

		materials=Sketchup.active_model.materials
		materials.each do |m|
			dicts=m.attribute_dictionaries
			if dicts
				dicts.delete("indigo07_materials")
				dicts.delete("uvsets")
			end
		next
		end
	end

	
	
	def SkIndigo.purge_mat

		materials=Sketchup.active_model.materials
		m=materials.current
		if m
			dicts=m.attribute_dictionaries
			if dicts
				dicts.delete("indigo07_materials")
				dicts.delete("uvsets")
			end
		end
	end
	
	
	#####################################
	def SkIndigo.purge_model

		m=Sketchup.active_model
		dicts=m.attribute_dictionaries
		dicts.delete("indigo07_settings")
	end

	###########################copy a file to a different location
	def SkIndigo.copy_file(source,dest)

		begin
			in_file=File.new((source),"r")
			in_data=in_file.read
			in_file.close

			out_file=File.new((dest),"w")
			out_file.write(in_data)
			out_file.close
		rescue
			UI.messagebox("SkIndigo file IO error.")
		end
	end


	#########################Function to check if current view is a photomatch view
	def SkIndigo.photomatch?(view)

		pages=Sketchup.active_model.pages
		return false if pages.count==0
		page = pages.selected_page
		return false if page==nil	 #gets the currently selected page
		page_label=page.label
		name_length=page.name.length
		label_length=page_label.length

		if label_length>name_length	#if using photomatch, label length will be greater than name length
			if page_label[name_length..label_length]==" (2D)"
				cam1=page.camera
						cam2=view.camera
						if (cam1.aspect_ratio==cam2.aspect_ratio) and (cam1.focal_length==cam2.focal_length) and (cam1.eye==cam2.eye) and (cam1.direction==cam2.direction) and (cam1.image_width==cam2.image_width) and (cam1.up==cam2.up)
					return true #current view is photomatch view
				end
			end
		end

		return false
	end


	###########
	def SkIndigo.igm_rgb_gamma_convert(colour,in_gamma,out_gamma)

		in_gamma=out_gamma if not in_gamma
		return nil if not colour
		rgb=colour.split(" ")
		if rgb.length==3
			red=((rgb[0].to_f)**(in_gamma.to_f/out_gamma.to_f))
			green=((rgb[1].to_f)**(in_gamma.to_f/out_gamma.to_f))
			blue=((rgb[2].to_f)**(in_gamma.to_f/out_gamma.to_f))
			return [red,green,blue]
		else
			return nil
		end
	end


	###########
	def SkIndigo.set_instancing(entities,instancing)

		entities.each {|ent|
			e=nil
			if ent.class==Sketchup::ComponentInstance
				e=ent.definition
			elsif ent.class==Sketchup::Group
				e=ent.entities[0].parent
			end
			e.set_attribute('indigo07_settings','inst',instancing)
		}
	end


	#########convert euler rotation matrix to rotation quaternion
	def SkIndigo.get_rotation_quaternion(trans_old,trans_new)

		old_x_axis=trans_old.xaxis
		old_y_axis=trans_old.yaxis
		old_z_axis=trans_old.zaxis

		xaxis=trans_new.xaxis
		yaxis=trans_new.yaxis
		zaxis=trans_new.zaxis

		#get the euler angles

		bank=xaxis.angle_between(old_x_axis)
		heading=zaxis.angle_between(old_y_axis)
		attitude=yaxis.angle_between(old_z_axis)
		#p bank.degrees
		#p heading.degrees
		#p attitude.degrees

		c1=Math.cos(heading/2.0)
		c2=Math.cos(attitude/2.0)
		c3=Math.cos(bank/2.0)
		s1=Math.sin(heading/2.0)
		s2=Math.sin(attitude/2.0)
		s3=Math.sin(bank/2.0)

		#note: I modified this from the reference website and it might not work.
		#switched y and z, also changed minus to plus due to axis direction
		w=c1*c2*c3-s1*s2*s3
		x=s1*s2*c3+c1*c2*s3
		y=c1*s2*c3+s1*c2*s3
		z=s1*c2*c3+c1*s2*s3

		return [w,x,y,z]
	end

	#########convert euler rotation matrix to rotation quaternion
	def SkIndigo.get_rotation_quaternion_void(transformation)

		xaxis=transformation.xaxis
		yaxis=transformation.yaxis
		zaxis=transformation.zaxis

		#get the euler angles

		bank=xaxis.angle_between(Geom::Vector3d.new(1,0,0))
		heading=zaxis.angle_between(Geom::Vector3d.new(0,0,1))
		attitude=yaxis.angle_between(Geom::Vector3d.new(0,1,0))

		c1=Math.cos(heading/2.0)
		c2=Math.cos(attitude/2.0)
		c3=Math.cos(bank/2.0)
		s1=Math.sin(heading/2.0)
		s2=Math.sin(attitude/2.0)
		s3=Math.sin(bank/2.0)

		#note: I modified this from the reference website and it might not work.
		w=c1*c2*c3-s1*s2*s3
		x=s1*s2*c3+c1*c2*s3
		y=c1*s2*c3+s1*c2*s3
		z=s1*c2*c3+c1*s2*s3

		return [w,x,y,z]
	end

	#########represent a quaternion in terms of axis-angle
	def SkIndigo.quaternion_as_axis_angle(quaternion)

		q=quaternion
		#p q
		w=q[0]
		i=q[1]
		j=q[2]
		k=q[3]

		angle=Math.acos(w)*2
		x=i/(Math.sin(angle/2.0))
		y=j/(Math.sin(angle/2.0))
		z=k/(Math.sin(angle/2.0))

		return [angle,Geom::Vector3d.new(x,y,z)]
	end

	#######
	def SkIndigo.get_axis_angle(transformation)

		xaxis=transformation.xaxis.normalize
		yaxis=transformation.yaxis.normalize
		zaxis=transformation.zaxis.normalize

		#calculate the angle
		angle=Math.acos((xaxis.x+yaxis.y+zaxis.z-1.0)/2.0)

		#calculate the axis
		axis=Geom::Vector3d.new
		#axis.x=yaxis.z-zaxis.y
		#axis.y=zaxis.x-xaxis.z
		#axis.z=xaxis.y-yaxis.x
		axis.x=zaxis.y-yaxis.z
		axis.z=yaxis.x-xaxis.y
		axis.y=xaxis.z-zaxis.x

		axis.normalize!

		if axis.length==0.0
			axis=Geom::Vector3d.new(0.0,1.0,0.0)
			angle=0.0
		end

		return [axis,angle]
	end


	############ javascriptEscape ##############################################

	def SkIndigo.javascriptEscape(str)
		# See http://tools.ietf.org/html/rfc7159#section-7
		# Replace double-quote characters (") with backslash double quote (\").
		# Replace backslash characters with double backslash.
		# Note that due to 'capture references', we need to write '\\\\\\' to get two backslashes.
		return str.gsub(/\\/, '\\\\\\').gsub(/"/, '\"')
	end


	############# Set options for a drop-down list in an HTML page. #############
	# web_dialog: a SketchUp API Web Dialog.
	# element_id should be a string
	# options_hash should be an array of 2-element arrays of strings(pairs of strings).
	# 
	# For example: SkIndigo.set_dropdown_options(@mat_editor, "MaterialName", [["paint", "paint"], ["glossy", "glossy"]])
	#
	def SkIndigo.set_dropdown_options(web_dialog, element_id, options_hash)

		options_json = "[" + options_hash.map do |keyvalue|
			key = keyvalue[0]
			value = keyvalue[1]
			escaped_key = javascriptEscape(key)
			escaped_value = javascriptEscape(value)
			
			"[\"#{escaped_key}\",\"#{escaped_value}\"]"
		end.join(",") + "]"
		

		#puts "============================"
		#puts "calling execute_script..."
		cmd = "setDropDownOptions(\"#{element_id}\", #{options_json})"
		#puts "cmd:"
		#puts cmd
		#puts "============================"
		t = Time.now
		web_dialog.execute_script(cmd)
		#puts "\t execute_script: #{Time.now - t} s"
	end
	
	
	def SkIndigo.set_value(dialog, id, value)
		cmd = "document.getElementById('#{id}').value=\"#{javascriptEscape(value.to_s)}\""
		dialog.execute_script(cmd)
	end
	
	def SkIndigo.set_checkbox(dialog, id, checked)
		cmd = "document.getElementById('#{id}').checked=#{checked ? "true" : "false"}"
		dialog.execute_script(cmd)
	end
	
	
	# This method exists to avoid errors like 'invalid byte sequence in UTF-8' when using e.g. the DMxml class.
	# See http://stackoverflow.com/questions/2982677/ruby-1-9-invalid-byte-sequence-in-utf-8
	def SkIndigo.remove_invalid_utf8_sequences(str)
		
		if String.method_defined?(:encode)
			str = str.encode('UTF-16', 'UTF-8', :invalid => :replace, :replace => '') # Convert to UTF-16
			str = str.encode('UTF-8', 'UTF-16') # Convert back to UTF-8
		end
		
		return str
	end
	
	def SkIndigo.has_prefix?(string, prefix)
		return string.slice(0, prefix.length) == prefix
	end
	
	def SkIndigo.remove_prefix(string, prefix)
		return string.slice(prefix.length, string.length - prefix.length)
	end
	
	# Returns full path if found, or nil if not found
	# See http://stackoverflow.com/questions/3498539/searching-a-folder-and-all-of-its-subfolders-for-files-of-a-certain-type
	def SkIndigo.find_file_in_dir_tree(dir_path, target_filename)
	
		dir_path_forwards_slashes = dir_path.gsub(/\\/, "/") # Backslashes don't work with Dir.glob().
	
		files = Dir.glob("#{dir_path_forwards_slashes}/**/#{target_filename}")
	
		return !files.empty? ? files[0] : nil
	end
	
	def SkIndigo.sanitise_string(str)
		return str.gsub(/[ \.\\\/:*#\[\]?"><\|]/, '_')
	end
	
	# See "What kinds of files can I import into SketchUp?", http://help.sketchup.com/en/article/36217
	def SkIndigo.openpanel_sketchup_supported_image(panel_caption)
		if SkIndigo.sketchup_is_2014_or_later # Multiple file type wildcards were broken before SketchUp 2014: http://www.sketchup.com/intl/en/developer/docs/ourdoc/ui?csw=1#openpanel
			filename = UI.openpanel(panel_caption, "", "Image file (*.bmp;*.jpg;*.jpeg;*.png;*.psd;*.tga;*.tif;*.tiff)|*.bmp;*.jpg;*.jpeg;*.png;*.psd;*.tga;*.tif;*.tiff||")
		else
			filename = UI.openpanel(panel_caption, "", "Image file (.bmp;.jpg;.jpeg;.png;.psd;.tga;.tif;.tiff)")
		end
	end

end ##########################################SkIndigo Module############################################






#---------------------------------------Custom Tool to select Focal distance----------------------
class IndigoFocusTool

	def initialize
		@ip1 = nil
			
	end

	def activate
		@ip1 = Sketchup::InputPoint.new
		@ip = Sketchup::InputPoint.new
		self.reset(nil)
	end

	def deactivate(view)
		view.invalidate 
	end

	def onMouseMove(flags, x, y, view)
				
		@ip.pick view, x, y
			if( @ip != @ip1 )
			   
				view.invalidate if( @ip.display? or @ip1.display? )
				@ip1.copy! @ip
							
				view.tooltip = @ip1.tooltip
			end

	end

	def onLButtonDown(flags, x, y, view)
			
			@ip1.pick view, x, y
			if( @ip1.valid? )

					model=Sketchup.active_model		
			model.set_attribute 'indigo07_settings','fp',@ip1.position
			UI.messagebox ("New Focal Point set to #{@ip1.position}")
								
			end
		
		view.lock_inference
		Sketchup.active_model.select_tool nil
	end


	def draw(view)
		if( @ip1.valid? )
			if( @ip1.display? )
				@ip1.draw(view)
				@drawn = true
			end
			
		end
	end

	def onCancel(flag, view)
		self.reset(view)
	end


	# Reset the tool back to its initial state
	def reset(view)
		   
		Sketchup::set_status_text("Select Point on Focal Plane", SB_PROMPT)
		@ip1.clear
			
		if( view )
			view.tooltip = nil
			view.invalidate 
		end
	  
	end

end # class FocusTool

####todo - put method inside skindigo module




####################Advanced user functions#####################
class Indigo

	def Indigo::adv

		indigo_path=SkIndigo.get_indigo_path
		return if indigo_path==nil
		inifile_path=File.dirname(indigo_path)
		inifile_path+="\/inifile.xml"
		if FileTest.exist?(inifile_path)
			UI.openURL(inifile_path)
		else
			puts "inifile.xml not found"
		end
	end

	def Indigo::ren

		indigo_path=SkIndigo.get_indigo_path
		return if indigo_path==nil
		renders_path=File.dirname(indigo_path)
		renders_path+="\/renders"
		if FileTest.exist?(renders_path)
			UI.openURL(renders_path) 
		else
			puts "renders directory not found"
		end
	end

	def Indigo::igs

		export_path=Sketchup.active_model.get_attribute 'indigo07_settings','export_path'
		return false if not export_path
		settings_path=File.join(File.dirname(export_path),(File.basename(export_path,".igs") + "-settings.igs"))
		if settings_path!=""
			if FileTest.exist?(settings_path)
				UI.openURL(settings_path)
			else
				puts "igs not found"
			end 
		else
			if FileTest.exist?(export_path)
				UI.openURL(export_path) 
			else
				puts "igs not found"
			end
		end
	end

	def Indigo::tm

		indigo_path=SkIndigo.get_indigo_path
		return if indigo_path==nil
		tm_path=File.dirname(indigo_path)
		tm_path+="\/IndigoToneMapper.exe"
		if FileTest.exist?(tm_path)
			UI.openURL(tm_path) 
		else
			puts "IndigoToneMapper.exe not found"
		end
	end

	def Indigo::web

		UI.openURL("http:\/\/www2.indigorenderer.com\/joomla\/")
	end

	def Indigo::mdb

		UI.openURL("http:\/\/www.elqx.com\/indmatdb\/")
	end


	def Indigo::scale=(val)
		Sketchup.active_model.set_attribute 'indigo07_settings','scale',val.to_f
	end

	def Indigo::scale
		Sketchup.active_model.get_attribute 'indigo07_settings','scale',1.0
	end

end #class Indigo end




###todo - put observer in separate file?
class SkindigoAppObserver <  Sketchup::AppObserver
	def onOpenModel(m)
		if $skindigo_mat_observer
			$skindigo_mat_observer.stop_preview()
			#timer=$skindigo_mat_observer.timer
			#UI.stop_timer(timer) if timer
			$skindigo_mat_observer.mat_editor.close if $skindigo_mat_observer.mat_editor.visible?
		end
		$skindigo_mat_observer=nil
		$skindigo_mat_observer=SkindigoMatObserver.new
		
		if $g_skindigo_settings
			$g_skindigo_settings.dialog.close if $g_skindigo_settings.dialog.visible?
		end
		$g_skindigo_settings=nil	
		
		if $g_skindigo_scatter_editor
			$g_skindigo_scatter_editor.dialog.close if $g_skindigo_scatter_editor.dialog.visible?
		end
		$g_skindigo_scatter_editor=nil	
		
		status=Sketchup.active_model.materials.add_observer($skindigo_mat_observer)
		$g_su2in_loaded_entities=nil if $g_su2in_loaded_entities
		if $skindigo_object_editor
			$skindigo_object_editor.dialog.close if $skindigo_object_editor.dialog.visible?
		end
		$skindigo_object_editor=nil
		$skindigo_object_editor=SkindigoObjectEditor.new
		Sketchup.active_model.selection.add_observer($skindigo_object_editor)
		IndigoMeshBuilder.clear_cached_meshes
		$skindigo_exported_this_session=false
	end

	def onNewModel(m)

		if $skindigo_mat_observer
			$skindigo_mat_observer.stop_preview()
			#timer=$skindigo_mat_observer.timer
			#UI.stop_timer(timer) if timer
			$skindigo_mat_observer.mat_editor.close if $skindigo_mat_observer.mat_editor.visible?
		end
		$skindigo_mat_observer=nil  
		$skindigo_mat_observer=SkindigoMatObserver.new
		
		if $g_skindigo_settings
			$g_skindigo_settings.dialog.close if $g_skindigo_settings.dialog.visible?
		end
		$g_skindigo_settings=nil
		
		if $g_skindigo_scatter_editor
			$g_skindigo_scatter_editor.dialog.close if $g_skindigo_scatter_editor.dialog.visible?
		end
		$g_skindigo_scatter_editor=nil
		
		Sketchup.active_model.materials.add_observer($skindigo_mat_observer)
		$g_su2in_loaded_entities=nil if $g_su2in_loaded_entities
		if $skindigo_object_editor
			$skindigo_object_editor.dialog.close if $skindigo_object_editor.dialog.visible?
		end
		$skindigo_object_editor=nil
		$skindigo_object_editor=SkindigoObjectEditor.new
		Sketchup.active_model.selection.add_observer($skindigo_object_editor)
		IndigoMeshBuilder.clear_cached_meshes
		$skindigo_exported_this_session=false
	end
	
	def onQuit()
		if $skindigo_mat_observer
			#UI.messagebox("test")
			$skindigo_mat_observer.stop_preview()
			Sketchup.active_model.materials.remove_observer($skindigo_mat_observer) 
			$skindigo_mat_observer=nil
		end
		$g_skindigo_settings=nil if $g_skindigo_settings
		$skindigo_object_editor=nil
		IndigoMeshBuilder.reset
		$skindigo_exported_this_session=nil
	end

end






#####
class SkIndigoFrameChangeObserver

	def initialize
		# @current page is used to make sure that the observer is only triggered by page changes
		# This is to workaround the PC bug that does not correctly populate the fromPage variable in the
		# callback to frameChange.
		@current_page = Sketchup.active_model.pages.selected_page
	end  # def initialize

	# Callback triggered by the Frame change observer.
	# fromPage is populated with the page we were coming from, toPage is
	# populated with the page we were transitioning to
	# and percent done contains the percentage of the transition that
	# has completed.

	def frameChange(fromPage, toPage, percent_done)      

		if @current_page != toPage then
			#update the current page
			#p "updating settings"
			@current_page = toPage
			$g_skindigo_settings.refresh_dialog(@current_page)

		end
	end   # def frameChange

end #class





######################
class IndigoColorPicker

	@@picker_name="SkIndigo Color Picker"

	def initialize(color)

		mats=Sketchup.active_model.materials
		picker_mat=mats[@@picker_name]

		if picker_mat
			@picker_mat=picker_mat
		else
			@picker_mat=mats.add(@@picker_name)
		end
		if @picker_mat
			@picker_mat.color=color
			mats.current=@picker_mat
		end
	end

	######
	def color

		return @picker_mat.color
	end

end  #end class color picker




################NKData class#####################
class NKdata

	def initialize(path)
		@path = path
	end

	def valid?
		return (@path and FileTest.exist?(self.expanded_path)) != false # This != false is actually required by ruby.
	end

	def valid_reference?
		path = @path.to_s
		return (FileTest.exist?(self.expanded_path) or (path=="")) != false
	end

	def expanded_path
		root = SkIndigo.get_resource_dir_path
		path = @path.to_s
		return File.join(root, "nkdata", File.basename(path))
	end

	def to_s
		return File.join("nkdata", File.basename(path)).gsub(/\\/, "\/")
	end

	def path
		return @path
	end

	def path=(new_path)
		@path = new_path
	end

	def filename
		File.basename(@path)
	end

	def name
		File.basename(@path, ".*")
	end

	def export(export_path)
		if not valid?
			UI.messagebox "NKdata path \"#{@path}\" is not valid."
			return
		else
			if FileTest.exist?(export_path)
				export_path = File.join(export_path, filename)
				nkdata = File.new(@path, "r")
				file_contents = nkdata.read
				nkdata.close
				export_file = File.new(export_path, "w")
				export_file.write(file_contents)
				export_file.close
			else
				UI.messagebox "Unable to write nkdata file to path #{export_path}"
			end
		end
	end
end #class NKdata





################IESProfile class#####################
class IESProfile < NKdata

	def expanded_path
		return @path
	end

	def to_s
		return @path.gsub(/\\/,"\/")
	end

	def export(export_path)
		if (not valid?)
			UI.messagebox "IES path is not valid."
			return
		else
			if FileTest.exist?(export_path)
				export_path=File.join(export_path,filename)
				ies_file=File.new((@path),"r")
				file_contents=ies_file.read
				ies_file.close
				export_file=File.new((export_path),"w")
				export_file.write(file_contents)
				export_file.close
			else
				UI.messagebox "Unable to write IES file to path #{export_path}"
			end
		end
	end
end #class IESProfile







#################################Texture Editor Class#############################
#todo put in separate file
class IndigoTextureEditor

	include SkIndigoUVS

	def initialize(texture,material,map_type=nil)   #tex is an IndigoTexture object - material is an IndigoMaterial object - map_type is abledo,bump,exponent,etc.
		@on_mac=SkIndigo.on_mac?
		@tex=texture
		#p @tex
		@mat=material
		@su_mat=@mat.mat
		@map_type=map_type
		@irs=IndigoRenderSettings.new
		@dlg=UI::WebDialog.new("SkIndigo Texture Editor", true, "SkIndigoTextureEditor",404,304,419,243,true) #width,height,left,top
		fn=File.join(File.dirname(reinterpret_as_utf8(__FILE__)),"TextureEditor.html")
		@dlg.set_file(fn)
		@temp_texture_path=ENV['temp']+'\texture_editor.jpg' if not @on_mac
		@temp_texture_path=File.join(ENV['TMPDIR'],"texture_editor.jpg") if @on_mac
		#p "temp texture path is #{@temp_texture_path}"
		@thumb=getThumbnail()
		#p "got thumbnail #{@thumb}"
		color = @dlg.get_default_dialog_color 
		@dlg.set_background_color(color)
		@dlg.add_action_callback("getValues") {|d,p|
			#p "Updating Values"
			getModelTextures()
			loadValuesFromSU()
			setVisibility()
		}
		@dlg.add_action_callback("load_uv_set") {|d,p|
			#uv_set=@dlg.get_element_value("uvset")
			#ents=Sketchup.active_model.selection
			#faces=ents.find_all {|e| e.class==Sketchup::Face}
			#if faces.length>0
			#	position_map(faces,uv_set)
			#end
			#defn=Sketchup.active_model.active_entities.parent
			#stored_set=defn.get_attribute('uvsets',uv_set)
			#if stored_set==true
				#UI.messagebox("UV set #{@tex.uv_set} loaded.")
			#else
			#	UI.messagebox("No UV data to load for UV set #{@tex.uv_set}.")
			#end
		}
		@dlg.add_action_callback("send_to_mapper") {|d,p|
			apply_settings(d)
			case @tex.type
				when "External"
					unless @tex.valid?
						UI.messagebox("You must first select a valid texture map.")
						break
					end
					if @tex.filename.slice(-3,3).downcase=="exr"
						UI.messagebox("EXR textures cannot be positioned.")
						break
					end
				when "SketchUp"
					if @su_mat.texture==nil
						UI.messagebox("You must first apply a texture to this material.")
						return
					end
				when "Constant"
					UI.messagebox("You must first load a texture.")
					return
			end
			uv_set=@tex.uv_set
			#if uv_set=="Default"
			#	result=UI.messagebox("By positioning this texture using the SketchUp UV set, you have chosen to apply this texture to the current material.  Continue?",MB_YESNO)
			#	if result==6 #yes
			#		@su_mat.texture=@temp_texture_path if @tex.type=="SketchUp"
			#		@su_mat.texture=@tex.path if @tex.type=="External"
			#		d.close()
			#		return
			#	else
			#		return
			#	end
			#end
			uv_mat=get_uv_mat()
			uv_mat.set_attribute('indigo_uv','current_set',@tex.uv_set.to_i)
			case @tex.type
				when "External"
					#uv_mat.texture=save_mat_texture(@su_mat,@temp_texture_path)  #store the currently applied texture of the material
					#@su_mat.texture=@tex.path
					uv_mat.texture=@tex.path
				when "SketchUp"
					uv_mat.texture=@temp_texture_path
					#uv_mat.texture=save_mat_texture(@su_mat,@temp_texture_path)
					#@su_mat.texture=@temp_texture_path
			end
			#ents=Sketchup.active_model.active_entities
			#faces=ents.find_all {|e| e.material==@su_mat}
			ents=Sketchup.active_model.selection
			faces=ents.find_all {|e| e.class==Sketchup::Face}
			if faces.length==0
				UI.messagebox("You must first select at least one face for positioning this texture.")
				return
			end
			Sketchup.active_model.start_operation("Position Map")
			#store_uvs(faces,0)  #save current uvs so they can be restored later
			switch_mat(faces,@mat.mat,uv_mat,@tex.uv_set.to_i)
			#try to load the stored uv set if the material has a texture
			#if uv_mat.texture!=nil
			#	position_map(faces,uv_set)
			#	$skindigo_current_uv_layer=uv_set.to_i
			#end
			d.close()
			defn=Sketchup.active_model.active_entities.parent
			stored_set=defn.get_attribute('uvsets',uv_set)
			#if stored_set==true
			#	UI.messagebox("UV set #{@tex.uv_set} loaded.")
			#else
			#	UI.messagebox("No UV data to load for UV set #{@tex.uv_set}.")
			#end
			Sketchup.active_model.commit_operation
		}
		@dlg.add_action_callback("apply_settings") {|d,p|
			apply_settings(d)
			$skindigo_mat_observer.refresh_dialog()
			d.close
		}
		@dlg.add_action_callback("update_thumbnail") {|d,p|
			textures=@irs.get_model_textures
			current_tex=d.get_element_value("TexturePath")
			new_thumb=textures.find {|path| path.gsub(/\\/,"\/")==current_tex}
			if new_thumb
				@thumb=new_thumb
				setThumbnail()
			end
		}
		@dlg.add_action_callback("cancel_changes") {|d,p| d.close}
		@dlg.add_action_callback("create_new_texture") {|d,p|
			if SkIndigo.sketchup_is_2014_or_later # Multiple file type wildcards were broken before SketchUp 2014: http://www.sketchup.com/intl/en/developer/docs/ourdoc/ui?csw=1#openpanel
				filename = UI.openpanel("Load New Texture", "", "Image file (*.bmp;*.exr;*.gif;*.hdr;*.jpg;*.jpeg;*.png;*.tga;*.tif;*.tiff)|*.bmp;*.exr;*.gif;*.hdr;*.jpg;*.jpeg;*.png;*.tga;*.tif;*.tiff||")
			else
				filename = UI.openpanel("Load New Texture", "", "Image file (.bmp;.exr;.gif;.hdr;.jpg;.jpeg;.png;.tga;.tif;.tiff)")
			end
			return if not filename
			
			case @tex.type
				when "SketchUp"
					@mat.mat.texture=filename
				when "External","Constant"
					@tex.path=filename
					@tex.type="External"
					@irs.store_texture(filename)
					#add_map_option(File.basename(filename))
					add_map_option(filename.gsub(/\\/,"\/"))
			end
			@thumb=getThumbnail()
			loadValuesFromSU()
			
		}
		@dlg.add_action_callback("refresh_thumbnail") {|d,p|
			@thumb=@dlg.get_element_value("TextureName")
			#p @thumb
		}
		#@dlg.show_modal {getModelTextures();loadValuesFromSU();setVisibility()}
		@dlg.show_modal {}
	end

	def apply_settings(d) #d is the dialog

		@tex.a=d.get_element_value("aValue")
		@tex.b=d.get_element_value("bValue")
		@tex.c=d.get_element_value("cValue")
		@tex.exponent=d.get_element_value("textureExponent")
		@tex.uv_set=d.get_element_value("uvset")
		path=d.get_element_value("TexturePath")
		smooth=d.get_element_value("smooth")
		if smooth=="true"
			@tex.smooth="On"
		else
			@tex.smooth="Off"
		end
		#p "setting texture path to #{path}"
		textures=@irs.get_model_textures
		current_tex=d.get_element_value("TexturePath")
		new_tex=textures.find {|path| path.gsub(/\\/,"\/")==current_tex}
		#p "new_tex set to #{new_tex}"
		@tex.path=new_tex

		#the map used to be applied by the material editor class.  however, this does not work on Mac because web dialogs are asynchronous.  We must apply the map inside this class for it to work on Mac
		case @map_type
		when "albedo"
			@mat.albedo_texture=@tex
		when "bump"
			@mat.bump_map=@tex
			@mat.use_bump=true if @tex.valid?
		when "normal"
			@mat.normal_map=@tex
			@mat.use_normal=true if @tex.valid?
		when "exponent"
			@mat.exponent_map=@tex
		when "displacement"
			@mat.disp_map=@tex
			@mat.use_displace=true if @tex.valid?
		when "clip"
			@mat.clip_map=@tex
			@mat.use_clip=true if @tex.valid?
		when "emission"
			@mat.emission_map=@tex
			@mat.use_emission=true if @tex.valid?
		when "transmittance"
			@mat.transmittance_map=@tex
		when "specular"
			@mat.specular_map=@tex
			@mat.use_specular=true if @tex.valid?
		when "blend"
			@mat.blend_map=@tex
		when "sigma"
			@mat.sigma_map=@tex
		when "roughness"
			@mat.roughness_map=@tex
		when "fresnel"
			@mat.fresnel_map=@tex
		when "thickness"
			@mat.thickness_map=@tex
		when "r_f"
			@mat.r_f_map=@tex
		when "front fresnel"
			@mat.front_fresnel_scale_map=@tex
		when "back fresnel"
			@mat.back_fresnel_scale_map=@tex
		when "front roughness"
			@mat.front_roughness_map=@tex
		when "back roughness"
			@mat.back_roughness_map=@tex
		when "abs_layer_trans"
			@mat.abs_layer_trans_map=@tex
			@mat.use_abs_layer_trans=true if @tex.valid?
		end
	end

	def show
		@dlg.show_modal
	end

	def setVisibility()
		type=@tex.type
		case type
			when "Constant","SketchUp"
				@dlg.execute_script("setSketchUpVisible()")
			when "External"
				@dlg.execute_script("setExternalVisible()")
		end
		case @map_type
			when "bump","exponent","displacement","clip","blend","sigma"
				@dlg.execute_script("$('#smoothTableData').attr('disabled', false);")
			else
				@dlg.execute_script("$('#smoothTableData').attr('disabled', true);")
		end
	end

	def getModelTextures()

		@model_textures=@irs.get_model_textures
		@model_textures.each {|path|
			add_map_option(path)
			#add_map_option(File.basename(path))
		}
	end

	def add_map_option(path)
		path=path.gsub(/\\/,"\/")
		cmd="createMapOption(\"TexturePath\",\"#{path}\",\"#{path}\")"
		@dlg.execute_script(cmd)
	end

	def loadValuesFromSU()
		#p "getting values from su"
		tex_path=@tex.path.gsub(/\\/,"\/")
		p "setting texture path #{tex_path}"
		setValue("TexturePath",tex_path) if @tex.type=="External"
		setValue("TextureName",@mat.display_name) if @tex.type=="SketchUp"
		setValue("TextureName","") if @tex.type=="Constant"
		setValue("aValue",@tex.a)
		setValue("bValue",@tex.b)
		setValue("cValue",@tex.c)
		setValue("textureExponent",@tex.exponent)
		setValue("uvset",@tex.uv_set)
		if (@tex.smooth=="On") then setCheckbox("smooth","true")
			else setCheckbox("smooth","false")
		end
		setThumbnail()
		
	end

	def setThumbnail()

		if @thumb
			@thumb.gsub!(/\\/,"\/")
			#cmd="document.images[\"TextureImage\"].src=\"nopreview.png\""  #this is required so the preview is refreshed
			#@dlg.execute_script(cmd)
			#p "setting thumbnail #{@thumb}"
			cmd="document.images[\"TextureImage\"].src=\"file:\/\/#{@thumb}?#{(rand*10000).to_i}\""
			@dlg.execute_script(cmd)
		end
	end

	def setValue(id, value)
		SkIndigo.set_value(@dlg, id, value)
	end

	def getThumbnail()
		case @tex.type
			when "Constant"
				return nil
			when "External"
				return @tex.path
			when "SketchUp"
				mat=@mat.mat  
				status=nil
				thumb=@temp_texture_path   #set the exported texture path to the user's temp directory
				#p thumb
				return save_mat_texture(mat,thumb)
		end
	end

	def setCheckbox(id,checked)
		cmd="document.getElementById('#{id}').checked=#{checked}"
		@dlg.execute_script(cmd)
	end

end #indigo texture editor class



	
	
	
#################################Shader Editor Class#############################
#todo - put in separate file
class IndigoShaderEditor

	def initialize(texture,material,map_type=nil)   #tex is an IndigoTexture object - material is an IndigoMaterial object
		@tex=texture
		@mat=material
		@map_type=map_type
		@dlg=UI::WebDialog.new("SkIndigo Shader Editor", true, "SkIndigoShaderEditor",630,350,419,243,true) #width,height,left,top
		@dlg.set_file(File.join(File.dirname(reinterpret_as_utf8(__FILE__)),"ShaderEditor.html"))
		color = @dlg.get_default_dialog_color 
		@dlg.set_background_color(color)
		@dlg.add_action_callback("apply_settings") {|d,p|
			apply_settings(d)
			#@tex.shader_code=d.get_element_value("shaderCode")
			d.close
		}
		@dlg.add_action_callback("cancel_changes") {|d,p| d.close}
		@dlg.show_modal {loadValuesFromSU()}
	end

	def show
		@dlg.show_modal
	end

	def loadValuesFromSU()
		code=@tex.shader_code
		#the following substitutions are necessary due to a strange issue with html textarea elements and newline characters
		code.gsub!(/\n/,"\\n")
		code.gsub!(/\r/,"\\r")
		setValue("shaderCode",code)
	end

	def setValue(id,value)
		new_value=value.to_s
		cmd="document.getElementById('#{id}').value=\"#{new_value}\""
		@dlg.execute_script(cmd)
	end

	def apply_settings(d) #d is the dialog

		@tex.shader_code=d.get_element_value("shaderCode")
		#the map used to be applied by the material editor class.  however, this does not work on Mac because web dialogs are asynchronous.  We must apply the map inside this class for it to work on Mac
		case @map_type
		when "albedo"
			@mat.albedo_texture=@tex
		when "bump"
			@mat.bump_map=@tex
		when "normal"
			@mat.normal_map=@tex
		when "exponent"
			@mat.exponent_map=@tex
		when "displacement"
			@mat.disp_map=@tex
		when "clip"
			@mat.clip_map=@tex
		when "emission"
			@mat.emission_map=@tex
		when "specular"
			@mat.specular_map=@tex
		when "blend"
			@mat.blend_map=@tex
		when "sigma"
			@mat.sigma_map=@tex
		when "abs_layer_trans"
			@mat.abs_layer_trans_map=@tex
		when "absorption"
			@mat.absorption_map=@tex
		when "sss"
			@mat.sss_map=@tex
		end
	end

end #indigo shader editor class

	
	
	



#################################Color Picker Class#############################

class ColorPicker

	def initialize(sucolor,indigo_material,property)   #color is a Sketchup color object
		
		@im=indigo_material
		@prop=property
		@rgb=sucolor
		@dlg=UI::WebDialog.new("Color Picker", true, "SkIndigoColorPicker",375,241,419,243,true) #width,height,left,top
		@dlg.set_file(File.join(File.dirname(reinterpret_as_utf8(__FILE__)),"colorpicker.html"))
		color = @dlg.get_default_dialog_color 
		@dlg.set_background_color(color)
		@dlg.add_action_callback("apply") {|d,p|
			
			r=@dlg.get_element_value("red")
			g=@dlg.get_element_value("green")
			b=@dlg.get_element_value("blue")
			@rgb=[r.to_i,g.to_i,b.to_i]
			#p @rgb
			case @prop
				when "base_emission"
					@im.base_emission=@rgb
			end
			$skindigo_mat_observer.refresh_dialog
			d.close
		}
		@dlg.add_action_callback("send_color") {|d,p|
			
		}
		@dlg.add_action_callback("set_color") {|d,p|
			setColor(@rgb)
		}
		@dlg.add_action_callback("cancel") {|d,p| d.close}
		@dlg.show_modal {}
		
	end

	def show
		@dlg.show_modal{}
	end

	def setColor(sucolor)  #rucolor is an rgb triplet array
		r=sucolor[0]
		b=sucolor[2]
		g=sucolor[1]
		cmd="$('#colorpicker').ColorPickerSetColor({r:#{r},g:#{g},b:#{b}});"
		@dlg.execute_script(cmd)
	end

	def getColor()
		@rgb
	end

end #color picker
	
	
	
	


class Sketchup::ComponentDefinition

	def indigo_instance?
		#return true
		instanced=self.get_attribute 'indigo07_settings','inst',false	
		return instanced
	end

end





#####################################Region Selection Tool####################################
class SelectRegionTool

	def initialize
		@ip1 = nil
		@ip2 = nil
		@xdown = 0
		@ydown = 0
		@x1=0
		@x2=0
		@y1=0
		@y2=0
	end

	def activate
		@ip1 = Sketchup::InputPoint.new
		@ip2 = Sketchup::InputPoint.new
		@ip = Sketchup::InputPoint.new
		@drawn = false    
		self.reset(nil)
	end

	def deactivate(view)
		view.invalidate if @drawn
	end

	def onMouseMove(flags, x, y, view)
		if( @state == 0 )
			# We are getting the first point of the region.  Call the pick method
			# on the InputPoint to get a 3D position from the 2D screen position
			# that is bassed as an argument to this method.
			@ip.pick view, x, y
			if( @ip != @ip1 )
				# if the point has changed from the last one we got, then
				# see if we need to display the point.  We need to display it
				# if it has a display representation or if the previous point
				# was displayed.  The invalidate method on the view is used
				# to tell the view that something has changed so that you need
				# to refresh the view.
				view.invalidate if( @ip.display? or @ip1.display? )
				@ip1.copy! @ip
			@x1=x
			@y1=y
				view.tooltip = @ip1.tooltip
			end
		else
			# Getting the second point of the region
			@ip2.pick view, x, y
			view.tooltip = @ip2.tooltip if( @ip2.valid? )
			view.invalidate
			
			# Update the length displayed in the VCB
			if( @ip2.valid? )
				length = @ip1.position.distance(@ip2.position)
				Sketchup::set_status_text length.to_s, SB_VCB_VALUE
			@x2=x
			@y2=y
			end
			
			# Check to see if the mouse was moved far enough to create a line.
			# This is used so that you can create a line by either draggin
			# or doing click-move-click
			if( (x-@xdown).abs > 10 || (y-@ydown).abs > 10 )
				@dragging = true
			end
		end
	end

	# The onLButtonDOwn method is called when the user presses the left mouse button.
	def onLButtonDown(flags, x, y, view)
		# When the user clicks the first time, we switch to getting the
		# second point.  When they click a second time we create the line
		if( @state == 0 )
			@ip1.pick view, x, y
			if( @ip1.valid? )
				@state = 1
				Sketchup::set_status_text("Select region", SB_PROMPT)
				@xdown = x
				@ydown = y
			@x1=x
			@y1=y
			end
		else
			# create the line on the second click
			if( @ip2.valid? )
				self.save_region(view)
				self.reset(view)
			end
		end
		
		# Clear any inference lock
		view.lock_inference
	end

	# The onLButtonUp method is called when the user releases the left mouse button.
	def onLButtonUp(flags, x, y, view)
		# If we are doing a drag, then create the line on the mouse up event
		if( @dragging && @ip2.valid? )
			self.save_region(view)
			self.reset(view)
		end
	end

	# onKeyDown is called when the user presses a key on the keyboard.
	# We are checking it here to see if the user pressed the shift key
	# so that we can do inference locking
	def onKeyDown(key, repeat, flags, view)
		if( key == CONSTRAIN_MODIFIER_KEY && repeat == 1 )
			@shift_down_time = Time.now
			
			# if we already have an inference lock, then unlock it
			if( view.inference_locked? )
				# calling lock_inference with no arguments actually unlocks
				view.lock_inference
			elsif( @state == 0 && @ip1.valid? )
				view.lock_inference @ip1
			elsif( @state == 1 && @ip2.valid? )
				view.lock_inference @ip2, @ip1
			end
		end
	end

	# onKeyUp is called when the user releases the key
	# We use this to unlock the interence
	# If the user holds down the shift key for more than 1/2 second, then we
	# unlock the inference on the release.  Otherwise, the user presses shift
	# once to lock and a second time to unlock.
	def onKeyUp(key, repeat, flags, view)
		if( key == CONSTRAIN_MODIFIER_KEY &&
			view.inference_locked? &&
			(Time.now - @shift_down_time) > 0.5 )
			view.lock_inference
		end
	end


	# The draw method is called whenever the view is refreshed.  It lets the
	# tool draw any temporary geometry that it needs to.
	def draw(view)
		if( @ip1.valid? )
			if( @ip1.display? )
				@ip1.draw(view)
				@drawn = true
			end
			
			if( @ip2.valid? )
				@ip2.draw(view) if( @ip2.display? )
				
				# The set_color_from_line method determines what color
				# to use to draw a line based on its direction.  For example
				# red, green or blue.
				view.set_color_from_line(@ip1, @ip2)
				self.draw_geometry([@x1,@y1,0],[@x2,@y1,0],[@x2,@y2,0],[@x1,@y2,0], view)
				@drawn = true
			end
		end
	end

	# onCancel is called when the user hits the escape key
	def onCancel(flag, view)
		self.reset(view)
	end


	# The following methods are not directly called from SketchUp.  They are
	# internal methods that are used to support the other methods in this class.

	# Reset the tool back to its initial state
	def reset(view)
		# This variable keeps track of which point we are currently getting
		@state = 0
		@x1=0
		@x2=0
		@y1=0
		@y2=0
		
		# Display a prompt on the status bar
		Sketchup::set_status_text("Select region", SB_PROMPT)
		
		# clear the InputPoints
		@ip1.clear
		@ip2.clear
		
		if( view )
			view.tooltip = nil
			view.invalidate if @drawn
		end
		
		@drawn = false
		@dragging = false
	end

	# Create new geometry when the user has selected two points.
	def save_region(view)

		irs=IndigoRenderSettings.new
		w=view.vpwidth.to_f
		h=view.vpheight.to_f
		width=irs.width
		height=irs.height
		cam=view.camera
		aspect=cam.aspect_ratio
		view_aspect=w/h
		aspect=view_aspect if aspect==0.0
		if view_aspect>=aspect   #black borders on left and right edges
			eff_width_pixels=aspect*h
			border_offset=((w-eff_width_pixels)/2).to_i
			width_fraction_x1=(@x1-border_offset)/eff_width_pixels
			width_fraction_x2=(@x2-border_offset)/eff_width_pixels
			height_fraction_y1=(@y1)/h
			height_fraction_y2=(@y2)/h
			x1=(width*width_fraction_x1).to_i
			x2=(width*width_fraction_x2).to_i
			y1=(height*height_fraction_y1).to_i
			y2=(height*height_fraction_y2).to_i

		elsif view_aspect<aspect  #black borders on top and bottom edges
			eff_height_pixels=w/aspect
			border_offset=((h-eff_height_pixels)/2).to_i
			width_fraction_x1=(@x1)/w
			width_fraction_x2=(@x2)/w
			height_fraction_y1=(@y1-border_offset)/eff_height_pixels
			height_fraction_y2=(@y2-border_offset)/eff_height_pixels
			x1=(width*width_fraction_x1).to_i
			x2=(width*width_fraction_x2).to_i
			y1=(height*height_fraction_y1).to_i
			y2=(height*height_fraction_y2).to_i
			
		end

		#Validate region and store

		min_x=[x1,x2].min
		max_x=[x1,x2].max
		min_y=[y1,y2].min
		max_y=[y1,y2].max
		min_x=0 if min_x<0
		max_x=width if max_x>width
		min_y=0 if min_y<0
		max_y=height if max_y>height
		reg=[min_x,max_x,min_y,max_y]
		UI.messagebox "Region (#{min_x},#{min_y}) (#{max_x},#{max_y}) selected."
		irs.region=reg
			
	end

	# Draw the geometry
	def draw_geometry(pt1, pt2, pt3, pt4, view)
		view.draw2d(GL_LINE_LOOP,[pt1,pt2,pt3,pt4])
	end

end #class




	

#-----------------------------------------------------------------------------
# This functions is just a shortcut for selecting the new tool
def select_region_tool
	Sketchup.active_model.select_tool SelectRegionTool.new
end


# ---------------------------------------- Menu item
if( not file_loaded?(__FILE__) )

	if ENV['SKINDIGO_SHOW_CONSOLE']
		Sketchup.send_action('showRubyPanel:') # open the Ruby Console - for development only.
	end

	main_menu = UI.menu("Plugins").add_submenu("SkIndigo")
	main_menu.add_item("Render Scene") {(SkIndigo.export(:normal, true, nil, true, false))}
	main_menu.add_separator
	main_menu.add_item("Export Scene as..") { (SkIndigo.export_as(:normal, true)) }
	main_menu.add_item("Export Sketchy Replay Animation") { (SkIndigo.export_animation(true)) }
	main_menu.add_item("Export Scene Tab Animation") { (SkIndigo.export_animation(false)) }
	main_menu.add_separator
	main_menu.add_item("Render Settings") { (SkIndigo.render_settings)}
	main_menu.add_item("Material Editor") { ($skindigo_mat_observer.show_editor if $skindigo_mat_observer) }
	main_menu.add_item("Query Material Types") { (SkIndigo.mat_query) }
	main_menu.add_item("Scatter Editor") { SkIndigo.show_scatter_editor }
	main_menu.add_separator
	main_menu.add_item("Set Light Layer Names") { SkIndigo.set_layer_names()}
	main_menu.add_item("Set Indigo Path") {SkIndigo.set_indigo_path}
	#main_menu.add_item("Resume a Render") { (Indigo.res) }
	main_menu.add_separator
	main_menu.add_item("About SkIndigo #{SkIndigo::VERSION}") { (SkIndigo.about) }
	
	Sketchup.add_observer(SkindigoAppObserver.new)

	$skindigo_mat_observer=SkindigoMatObserver.new
	Sketchup.active_model.materials.add_observer $skindigo_mat_observer
	
	$skindigo_object_editor=SkindigoObjectEditor.new
	Sketchup.active_model.selection.add_observer $skindigo_object_editor

	plugins = Sketchup.find_support_file("Plugins\/skindigo")
	icon1 = File.join(plugins,"IndigoLogo.gif")
	icon2 = File.join(plugins, "IndigoLogo3.gif")
	icon_3new = File.join(plugins, "ExportSelected.png") # Export selected entities only.
	icon3 = File.join(plugins, "IndigoLogo4.gif")
	icon4 = File.join(plugins, "IndigoLogo2.gif")
		tb = UI::Toolbar.new("SkIndigo")
		
	cmd1 = UI::Command.new("Render Scene"){(SkIndigo.export(:normal, true, nil, true, 
		false # Export selected entities only
	))} 
	cmd1.small_icon = icon1
	cmd1.large_icon = icon1
	cmd1.tooltip = cmd1.status_bar_text = "Render with Indigo"
	
	
	cmd3_new = UI::Command.new("Render Selected Entities"){(SkIndigo.export(:normal, true, nil, true,
		true # Export selected entities only
	))}
	cmd3_new.menu_text="   Render Selected Entities"
	cmd3_new.set_validation_proc { SkIndigo.validate_export_selected_entities_only }
	cmd3_new.small_icon = icon_3new
	cmd3_new.large_icon = icon_3new
	cmd3_new.tooltip = cmd3_new.status_bar_text = "Render Selected Entities"
	
	cmd3 = UI::Command.new("Materal Editor"){($skindigo_mat_observer.show_editor if $skindigo_mat_observer)}
	cmd3.small_icon = icon3
	cmd3.large_icon = icon3
	cmd3.tooltip = cmd3.status_bar_text = "Material Editor"
	
	cmd4 = UI::Command.new("Render Settings"){(SkIndigo.render_settings)}
	cmd4.menu_text="Render Settings"
	cmd4.small_icon = icon4
	cmd4.large_icon = icon4
	cmd4.tooltip = cmd4.status_bar_text = "Render Settings"
	
	cmd5 = UI::Command.new("Object Editor"){($skindigo_object_editor.show if $skindigo_object_editor)}
	cmd5.menu_text="Object Editor"
	cmd5.small_icon=icon1
	cmd5.large_icon=icon1
	
	tb.add_item(cmd1)
	tb.add_item(cmd3_new)
	tb.add_item(cmd3)
	tb.add_item(cmd4)
	#tb.add_item(cmd5)
	tb.show
		
		
	##############################UV Manager Context menu##########################
	
	UI.add_context_menu_handler do |menu|
		ent=Sketchup.active_model.selection.first
		mat=nil
		if ent
			if ent.material
				mat=IndigoMaterial.new(ent.material)
			elsif (ent.class==Sketchup::Face) and ent.back_material
				mat=IndigoMaterial.new(ent.back_material)
			end
		end
		menu.add_separator
		menu.add_item("SkIndigo"){(nil)}
		if mat
			unless mat.color_picker? or mat.uv_helper?
				menu.add_item("   Edit \"#{mat.name}\" (#{mat.type.upcase})") {($skindigo_mat_observer.sm=mat;$skindigo_mat_observer.show_editor)}
			end
		end
	end
	

	
	UI.add_context_menu_handler do |menu|
		sel=Sketchup.active_model.selection
		ent=sel.first
		if ent.class==Sketchup::SectionPlane
			unless (ent.get_attribute('indigo07_settings','active_section',false))
				menu.add_item("   Enable Section Plane") {ent.set_attribute('indigo07_settings','active_section',true)}
			else
				menu.add_item("   Disable Section Plane") {ent.set_attribute('indigo07_settings','active_section',false)}
			end
		end
		
	end #menu
	
	UI.add_context_menu_handler do |menu|
	
		menu.add_item("   Edit Active Mesh") {($skindigo_object_editor.show(Sketchup.active_model.active_entities.parent) if $skindigo_object_editor)}
		
		sel=Sketchup.active_model.selection
		ent=sel.first
		e=nil
		if ent.class==Sketchup::ComponentInstance
			e=ent.definition
		elsif ent.class==Sketchup::Group
			e=ent.entities[0].parent
		end
		
		if e and sel.length==1
			#instanced=e.get_attribute 'indigo07_settings','inst',false
			#if not instanced
				#menu.add_item("   \"#{e.name}\" Enable Instancing") {(e.set_attribute 'indigo07_settings','inst',true)}
			#else
				#menu.add_item("   \"#{e.name}\" Disable Instancing") {(e.set_attribute 'indigo07_settings','inst',false)}
				#menu.add_item("   Edit Object") {($skindigo_object_editor.show(e) if $skindigo_object_editor)}
				menu.add_item("   Set Keyframe") {SkIndigo.set_keyframe([ent])}
				menu.add_item("   Clear Keyframes") {SkIndigo.clear_keyframes([ent])} if ent.has_indigo_keyframes?
			#end
		elsif e and sel.length>1
			#menu.add_item("   Enable Instancing") {(SkIndigo.set_instancing(sel,true))}
			#menu.add_item("   Disable Instancing") {(SkIndigo.set_instancing(sel,false))}
			menu.add_item("   Set Keyframes") {SkIndigo.set_keyframe(sel)}
			menu.add_item("   Clear Keyframes") {SkIndigo.clear_keyframes(sel)}
		end	
	end
	

	SkIndigo.check_indigo_version
	
	
	# Copied from preview_running? in SkindigoMatObserver
	def SkIndigo.indigo_gui_running?()
		wmi = WIN32OLE.connect("winmgmts://")
		query = "select * from Win32_process"
		processes = wmi.ExecQuery(query)
		names=[]
		processes.each{|p| names.push(p.Name)}
		return true if (names.include?("indigo.exe") || names.include?("indigo_gui.exe"))
	end #function
	
	if SkIndigo.run_tests?
		puts "================================================="
		puts "=========== Running SkIndigo tests =============="
		puts "================================================="
		
		# Get list of scenes to export from the SKINDIGO_SCENE_PATHS env var
		scenes = ENV["SKINDIGO_SCENE_PATHS"].split(";")
		
		begin
			scenes.each do |scene_path|
				
				puts "\n===Opening scene:====\n#'#{scene_path}'..."
			
				Sketchup.open_file(scene_path)
			
				#puts "Exporting..."
				result = SkIndigo.export(
					:normal, # mode
					true, # textures
					nil, # export_path
					true, # auto_render
					false # should_export_selected_entities_only
				)
				raise "Export of scene '#{scene_path}' failed." if !result
				
				sleep(0.1)
				while SkIndigo.indigo_gui_running?
					sleep(0.1)
					#puts "Waiting...."
				end
			end
			puts "=================================================="
			puts "============= Done running tests. ================"
			puts "=================================================="
		rescue
			puts "==================================================="
			puts "Exception thrown: #{$!}"
			puts "==================================================="
		end
	end

end 


#-----------------------------------------------------------
file_loaded(__FILE__)

